"""
The setting of Superb KS

Authors
  * Yist Y. Lin 2021
  * Leo 2022
"""

import logging
import pickle
from pathlib import Path
from typing import OrderedDict

import pandas as pd
from omegaconf import MISSING
from torch.utils.data import Dataset

from s3prl.dataio.corpus.speech_commands import SpeechCommandsV1
from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.dataio.sampler import BalancedWeightedSampler, FixedBatchSizeBatchSampler
from s3prl.nn.linear import MeanPoolingLinear

from .superb_sid import SuperbSID

logger = logging.getLogger(__name__)


__all__ = [
    "gsc1_for_classification",
    "SuperbKS",
]


def gsc1_for_classification(
    target_dir: str,
    cache_dir: str,
    gsc1: str,
    gsc1_test: str,
    get_path_only: bool = False,
):
    """
    Prepare Google Speech Command for classfication task
    following :obj:`SuperbKS.prepare_data` format.

    Args:
        gsc1 (str): The root path of the Google Speech Command V1 training set
        gsc1_test (str): The root path of the Google Speech Command V1 test set
        **others: refer to :obj:`SuperbKS.prepare_data`
    """
    target_dir = Path(target_dir)

    train_path = target_dir / f"train.csv"
    valid_path = target_dir / f"valid.csv"
    test_paths = [target_dir / f"test.csv"]

    if get_path_only:
        return train_path, valid_path, test_paths

    def gsc_v1_for_superb(gsc1: str, gsc1_test: str):
        corpus = SpeechCommandsV1(gsc1, gsc1_test)

        def format_fields(data: dict):
            import torchaudio

            formated_data = OrderedDict()
            for key, value in data.items():
                data_point = {
                    "wav_path": value["wav_path"],
                    "label": value["class_name"],
                    "start_sec": None,
                    "end_sec": None,
                }
                if value["class_name"] == "_silence_":
                    # NOTE: for silence, crop into 1-second segments, which
                    # is the standard way reported in the original paper
                    info = torchaudio.info(value["wav_path"])
                    for start in list(range(info.num_frames))[:: info.sample_rate]:
                        seg = data_point.copy()
                        end = min(start + 1 * info.sample_rate, info.num_frames)
                        seg["start_sec"] = start / info.sample_rate
                        seg["end_sec"] = end / info.sample_rate
                        formated_data[f"{key}_{start}_{end}"] = seg
                else:
                    formated_data[key] = data_point
            return formated_data

        train_data, valid_data, test_data = corpus.data_split
        return (
            format_fields(train_data),
            format_fields(valid_data),
            format_fields(test_data),
        )

    train_data, valid_data, test_data = gsc_v1_for_superb(gsc1, gsc1_test)

    def dict_to_csv(data_dict, csv_path):
        keys = sorted(list(data_dict.keys()))
        fields = sorted(data_dict[keys[0]].keys())
        data = dict()
        for field in fields:
            data[field] = []
            for key in keys:
                data[field].append(data_dict[key][field])
        data["id"] = keys
        df = pd.DataFrame(data)
        df.to_csv(csv_path, index=False)

    dict_to_csv(train_data, train_path)
    dict_to_csv(valid_data, valid_path)
    dict_to_csv(test_data, test_paths[0])

    return train_path, valid_path, test_paths


class SuperbKS(SuperbSID):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                gsc1=MISSING,
                gsc1_test=MISSING,
            ),
            build_encoder=dict(),
            build_dataset=dict(
                train=dict(
                    sox_effects=[
                        ["channels", "1"],
                        ["rate", "16000"],
                        ["gain", "-3.0"],
                    ],
                ),
                valid=dict(
                    sox_effects=[
                        ["channels", "1"],
                        ["rate", "16000"],
                        ["gain", "-3.0"],
                    ],
                ),
                test=dict(
                    sox_effects=[
                        ["channels", "1"],
                        ["rate", "16000"],
                        ["gain", "-3.0"],
                    ],
                ),
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                ),
                valid=dict(
                    batch_size=32,
                ),
                test=dict(
                    batch_size=32,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=256,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-4,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=200000,
                log_step=100,
                eval_step=5000,
                save_step=1000,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="accuracy",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        """
        Prepare the task-specific data metadata (path, labels...).
        By default call :obj:`gsc1_for_classification` with :code:`**prepare_data`

        Args:
            prepare_data (dict): same in :obj:`default_config`,
                support arguments in :obj:`gsc1_for_classification`
            target_dir (str): Parse your corpus and save the csv file into this directory
            cache_dir (str): If the parsing or preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            tuple

            1. train_path (str)
            2. valid_path (str)
            3. test_paths (List[str])

            Each path (str) should be a csv file containing the following columns:

            ====================  ====================
            column                description
            ====================  ====================
            id                    (str) - the unique id for this data point
            wav_path              (str) - the absolute path of the waveform file
            label                 (str) - a string label of the waveform
            start_sec             (float) - optional, load the waveform from :code:`start_sec` seconds. If not presented or is :code:`math.nan`, load from the beginning.
            end_sec               (float) - optional, load the waveform from :code:`end_sec` seconds. If not presented or is :code:`math.nan`, load to the end.
            ====================  ====================
        """
        return gsc1_for_classification(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_encoder(
        self,
        build_encoder: dict,
        target_dir: str,
        cache_dir: str,
        train_csv_path: str,
        valid_csv_path: str,
        test_csv_paths: list,
        get_path_only: bool = False,
    ):
        """
        Build the encoder (for the labels) given the data metadata, and return the saved encoder path.
        By default generate and save a :obj:`s3prl.dataio.encoder.CategoryEncoder` from the :code:`label` column of all the csv files.

        Args:
            build_encoder (dict): same in :obj:`default_config`, no argument supported for now
            target_dir (str): Save your encoder into this directory
            cache_dir (str): If the preprocessing takes too long time, you can save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            train_csv_path (str): the train path from :obj:`prepare_data`
            valid_csv_path (str): the valid path from :obj:`prepare_data`
            test_csv_paths (List[str]): the test paths from :obj:`prepare_data`
            get_path_only (str): Directly return the filepaths no matter they exist or not.

        Returns:
            str

            tokenizer_path: The tokenizer should be saved in the pickle format
        """
        encoder_path = Path(target_dir) / "encoder.pkl"
        if get_path_only:
            return encoder_path

        train_csv = pd.read_csv(train_csv_path)
        valid_csv = pd.read_csv(valid_csv_path)
        test_csvs = [pd.read_csv(path) for path in test_csv_paths]
        all_csv = pd.concat([train_csv, valid_csv, *test_csvs])

        labels = all_csv["label"].tolist()
        encoder = CategoryEncoder(labels)
        with open(encoder_path, "wb") as f:
            pickle.dump(encoder, f)

        return encoder

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset: Dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.
        By default for train and valid, use :obj:`BalancedWeightedSampler`; for test use
        :obj:`FixedBatchSizeBatchSampler`

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`BalancedWeightedSampler`
                valid                 (dict) - arguments for :obj:`BalancedWeightedSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        def _build_batch_sampler(
            train: dict = None, valid: dict = None, test: dict = None
        ):
            train = train or {}
            valid = valid or {}
            test = test or {}

            csv = pd.read_csv(data_csv)
            labels = csv["label"].tolist()

            if mode == "train":
                return BalancedWeightedSampler(labels, **train)
            elif mode == "valid":
                return BalancedWeightedSampler(labels, **valid)
            elif mode == "test":
                return FixedBatchSizeBatchSampler(csv, **test)

        return _build_batch_sampler(**build_batch_sampler)

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_downsample_rate: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`MeanPoolingLinear` model

        Args:
            build_downstream (dict): same in :obj:`default_config`,
                support arguments of :obj:`MeanPoolingLinear`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`AbsUtteranceModel`
        """
        model = MeanPoolingLinear(
            downstream_input_size, downstream_output_size, **build_downstream
        )
        return model



"""
Wrap any batch sampler for distributed training

Authors:
  * Leo 2022
"""

import logging
from copy import deepcopy
from typing import Iterator, Optional, TypeVar

import torch.distributed as dist
from torch.utils.data import BatchSampler

T_co = TypeVar("T_co", covariant=True)
logger = logging.getLogger(__name__)

__all__ = [
    "DistributedBatchSamplerWrapper",
]


class DistributedBatchSamplerWrapper:
    def __init__(
        self,
        batch_sampler: BatchSampler,
        num_replicas: Optional[int] = None,
        rank: Optional[int] = None,
        allow_duplicates: bool = False,
        allow_uneven: bool = False,
    ) -> None:
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        if rank >= num_replicas or rank < 0:
            raise ValueError(
                "Invalid rank {}, rank should be in the interval"
                " [0, {}]".format(rank, num_replicas - 1)
            )
        self.batch_sampler = batch_sampler
        self.num_replicas = num_replicas
        self.rank = rank
        self.allow_duplicates = allow_duplicates
        self.allow_uneven = allow_uneven

    def __iter__(self) -> Iterator[T_co]:
        logger.info(
            f"Building distributed batch sampler for rank={self.rank}, world_size={self.num_replicas}"
        )

        all_rank_batch_indices = list(iter(self.batch_sampler))
        if len(all_rank_batch_indices) % self.num_replicas == 0:
            target_batch_indices = all_rank_batch_indices
        else:
            num_to_halve = (
                self.num_replicas - len(all_rank_batch_indices) % self.num_replicas
            )
            flatten_batch_indices = deepcopy(all_rank_batch_indices)
            while num_to_halve > 0:
                newly_flatten = []
                all_cant_be_halved = True
                for indices in flatten_batch_indices:
                    if num_to_halve > 0 and len(indices) > 1:
                        indices1, indices2 = (
                            indices[: len(indices) // 2],
                            indices[len(indices) // 2 :],
                        )
                        newly_flatten += [indices1, indices2]
                        num_to_halve -= 1
                        all_cant_be_halved = False
                    else:
                        newly_flatten.append(indices)
                flatten_batch_indices = deepcopy(newly_flatten)

                if all_cant_be_halved:
                    if self.allow_duplicates:
                        logger.warning(
                            "To ensure all the dataloaders in different processes get the same number "
                            "of batches. Some batches are duplicated. This must not happen during the "
                            "evaluation stage."
                        )
                        flatten_batch_indices = (
                            flatten_batch_indices
                            + all_rank_batch_indices[:num_to_halve]
                        )
                    elif self.allow_uneven:
                        logger.warning(
                            "Total batches will not be evenly distributed across the dataloaders in "
                            "different processes. This must not happen during the training stage and "
                            "can lead to hanging, while might be okay during the evaluation stage."
                        )
                    else:
                        raise ValueError(
                            "The provided batch sampler cannot be safely wrapped for distributed training. "
                            "Please try increase the number of indices in each batch. Or, allowing duplicated "
                            "batches or uneven number of batches across dataloaders."
                        )
            target_batch_indices = flatten_batch_indices

        if not self.allow_uneven:
            assert len(target_batch_indices) % self.num_replicas == 0

        batch_indices = target_batch_indices[self.rank :: self.num_replicas]
        return iter(batch_indices)

    def __len__(self) -> int:
        # Since the total number of batches dynamically depends on the current epoch,
        # instead of pre-compute it which will duplicate the batch number computation logic,
        # it makes no harm to simply re-compute it with __iter__ for every call, since
        # __len__ is usually not frequently called and won't be the performance bottleneck
        return len(list(iter(self)))

    def set_epoch(self, epoch: int) -> None:
        if hasattr(self.batch_sampler, "set_epoch"):
            self.batch_sampler.set_epoch(epoch)



from dataclasses import dataclass

import torch

from .base import AugmentedDynamicItemDataset, DataPipe


@dataclass
class LabelMaskFromLen(DataPipe):
    target_feat_name: str = (
        "target_feat"  # tensors in the shape of: (seq_len, feat_dim)
    )
    label_mask_name: str = "label_mask"

    def create_label_mask(self, target_feat):
        with torch.no_grad():
            seq_len = target_feat.shape[0]
            label_mask = torch.ones_like(target_feat, dtype=torch.uint8)
            label_mask[seq_len:, :] = 0
            label_mask = label_mask.to(dtype=torch.bool)

        return label_mask

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.create_label_mask,
            takes=[self.target_feat_name],
            provides=[
                self.label_mask_name,
            ],
        )
        return dataset



"""
Speech2Text with CTC loss

Authors
  * Heng-Jui Chang 2022
"""

import logging
from pathlib import Path
from typing import List, Union

import numpy as np
import torch
import torch.nn.functional as F
from torch import nn

from s3prl.dataio.encoder.tokenizer import Tokenizer
from s3prl.metric import cer, per, wer
from s3prl.metric.slot_filling import (
    slot_edit_f1_full,
    slot_edit_f1_part,
    slot_type_f1,
    slot_value_cer,
    slot_value_wer,
)
from s3prl.nn import BeamDecoder

from . import Task

logger = logging.getLogger(__name__)

__all__ = [
    "Speech2TextCTCExample",
    "Speech2TextCTCTask",
]


class Speech2TextCTCExample(nn.Module):
    """An example speech-to-text task with CTC objective

    Args:
        input_size (int, optional): Input size. Defaults to 3.
        output_size (int, optional): Output size. Defaults to 4.
    """

    def __init__(self, input_size=3, output_size=4):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

    @property
    def input_size(self):
        return self._input_size

    @property
    def output_size(self):
        return self._output_size

    def forward(self, x, x_len):
        """
        Args:
            x (torch.Tensor): (batch_size, timestemps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            y (torch.Tensor): (batch_size, output_size)
            y_len (torch.LongTensor): (batch_size)
        """
        assert x.size(-1) == self.input_size
        output = torch.randn(x.size(0), x.size(1), self.output_size)
        assert output, x_len


class Speech2TextCTCTask(Task):
    """Speech-to-text task with CTC objective

    Args:
        model (Speech2TextCTCExample)
        tokenizer (Tokenizer): Text tokenizer.
        decoder (Union[BeamDecoder, dict], optional):
            Beam decoder or decoder's config. Defaults to None.
        log_metrics (List[str], optional):
            Metrics to be logged. Defaults to ["cer", "wer"].
    """

    def __init__(
        self,
        model: torch.nn.Module,
        tokenizer: Tokenizer,
        decoder: Union[BeamDecoder, dict] = None,
        log_metrics: List[str] = ["cer", "wer"],
    ) -> None:
        super().__init__()
        self.model = model

        assert isinstance(tokenizer, Tokenizer)
        self.tokenizer = tokenizer
        self.log_metrics = log_metrics

        if BeamDecoder is None:
            decoder = None
        if isinstance(decoder, dict):
            decoder = BeamDecoder(**decoder)
            logger.info("Using flashlight decoder.")
        self.decoder = decoder

        self.criterion = nn.CTCLoss(
            blank=self.tokenizer.pad_idx,
            zero_infinity=True,
        )

    def predict(self, x: torch.Tensor, x_len: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, timestamps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            logits (torch.Tensor): (batch_size, timestamps, output_size)
            prediction (list): prediction strings
            valid_length (torch.LongTensor): (batch_size, )
        """

        logits, x_len = self.model(x, x_len)
        predicted_tokens = torch.argmax(logits, dim=2).detach().cpu()
        filtered_tokens = [
            [
                token
                for token in pred_token.unique_consecutive().tolist()
                if token != self.tokenizer.pad_idx and token != self.tokenizer.eos_idx
            ]
            for pred_token in predicted_tokens
        ]
        predictions = [
            self.tokenizer.decode(token_list) for token_list in filtered_tokens
        ]
        return logits, predictions, x_len

    def forward(
        self,
        _mode: str,
        x: torch.Tensor,
        x_len: torch.LongTensor,
        labels: np.ndarray,
        class_ids: torch.LongTensor,
        unique_name: np.ndarray,
        beam_decode: bool = False,
        _dump_dir: str = None,
    ):
        """
        Each forward step in the training loop

        Args:
            mode (str): train / valid / test
            x (torch.Tensor):
                Input waveform or acoustic features.
                (batch_size, timestamps, input_size)
            x_len (torch.LongTensor):
                Lengths of inputs.
                (batch_size, )
            labels (np.ndarray):
                Ground truth transcriptions (str).
                (batch_size, )
            class_ids (torch.LongTensor):
                Tokenized ground truth transcriptions.
            unique_name (np.ndarray):
                Unique names for each sample.

        """
        logits, prediction, x_len = self.predict(x, x_len)
        log_probs = F.log_softmax(logits, dim=2)

        y = class_ids
        y_len = torch.tensor(
            [(ids != self.tokenizer.pad_idx).long().sum() for ids in class_ids],
            dtype=torch.long,
            device=logits.device,
        )

        loss = self.criterion(log_probs.transpose(0, 1), y, x_len, y_len)

        hyps = None
        if beam_decode and self.decoder is not None:
            hyps = self.decoder.decode(log_probs.detach())

        cacheable = dict(
            loss=loss.detach().cpu().item(),
            prediction=prediction,
            label=labels.tolist(),
            unique_name=unique_name.tolist(),
            hypotheses=hyps,
        )

        return loss, cacheable

    def reduction(self, _mode: str, cached_results: List[dict], _dump_dir: str = None):
        results = self.parse_cached_results(cached_results)

        losses = results["loss"]
        predictions = results["prediction"]
        labels = results["label"]
        unique_names = results["unique_name"]

        if _dump_dir is not None:
            with (Path(_dump_dir) / "hyp").open("w") as f:
                f.writelines(
                    [f"{uid} {p}\n" for p, uid in zip(predictions, unique_names)]
                )

            with (Path(_dump_dir) / "ref").open("w") as f:
                f.writelines([f"{uid} {p}\n" for p, uid in zip(labels, unique_names)])

        beam_hyps = None
        if results["hypotheses"][0] is not None:
            beam_hyps = [" ".join(hyp[0].words) for hyp in results["hypotheses"]]

        logs = {}
        logs["loss"] = float(np.mean(losses))

        if "wer" in self.log_metrics:
            logs["wer"] = wer(predictions, labels)
        if "cer" in self.log_metrics:
            logs["cer"] = cer(predictions, labels)
        if "per" in self.log_metrics:
            logs["per"] = per(predictions, labels)
        if "slot_type_f1" in self.log_metrics:
            logs["slot_type_f1"] = slot_type_f1(predictions, labels)
        if "slot_value_cer" in self.log_metrics:
            logs["slot_value_cer"] = slot_value_cer(predictions, labels)
        if "slot_value_wer" in self.log_metrics:
            logs["slot_value_wer"] = slot_value_wer(predictions, labels)
        if "slot_edit_f1_full" in self.log_metrics:
            logs["slot_edit_f1_full"] = slot_edit_f1_full(predictions, labels)
        if "slot_edit_f1_part" in self.log_metrics:
            logs["slot_edit_f1_part"] = slot_edit_f1_part(predictions, labels)

        if beam_hyps is not None:
            logs["wer_beam"] = wer(beam_hyps, labels)
            logs["char_beam"] = cer(beam_hyps, labels)

        return logs



import os
import re
from collections import Counter
from multiprocessing import Pool
from pathlib import Path

import torch

SPACE_NORMALIZER = re.compile(r"\s+")


def utils_item(tensor):
    # tpu-comment: making this a no-op for xla devices.
    if torch.is_tensor(tensor) and tensor.device.type == "xla":
        return tensor.detach()
    if hasattr(tensor, "item"):
        return tensor.item()
    if hasattr(tensor, "__getitem__"):
        return tensor[0]
    return tensor


def find_offsets(filename: str, num_chunks: int):
    """
    given a file and a number of chuncks, find the offsets in the file
    to be able to chunk around full lines.
    """
    with open(filename, "r", encoding="utf-8") as f:
        size = os.fstat(f.fileno()).st_size
        chunk_size = size // num_chunks
        offsets = [0 for _ in range(num_chunks + 1)]
        for i in range(1, num_chunks):
            f.seek(chunk_size * i)
            _safe_readline(f)
            offsets[i] = f.tell()
        offsets[-1] = size
        return offsets


def post_process(sentence: str, symbol: str):
    if symbol == "sentencepiece":
        sentence = sentence.replace(" ", "").replace("\u2581", " ").strip()
    elif symbol == "wordpiece":
        sentence = sentence.replace(" ", "").replace("_", " ").strip()
    elif symbol == "letter":
        sentence = sentence.replace(" ", "").replace("|", " ").strip()
    elif symbol == "silence":
        import re

        sentence = sentence.replace("<SIL>", "")
        sentence = re.sub(" +", " ", sentence).strip()
    elif symbol == "_EOW":
        sentence = sentence.replace(" ", "").replace("_EOW", " ").strip()
    elif symbol in {"subword_nmt", "@@ ", "@@"}:
        if symbol == "subword_nmt":
            symbol = "@@ "
        sentence = (sentence + " ").replace(symbol, "").rstrip()
    elif symbol == "none":
        pass
    elif symbol is not None:
        raise NotImplementedError(f"Unknown post_process option: {symbol}")
    return sentence


def _safe_readline(fd) -> str:
    pos = fd.tell()
    while True:
        try:
            return fd.readline()
        except UnicodeDecodeError:
            pos -= 1
            fd.seek(pos)  # search where this character begins


class ChunkLineIterator:
    """
    Iterator to properly iterate over lines of a file chunck.
    """

    def __init__(self, fd, start_offset: int, end_offset: int):
        self._fd = fd
        self._start_offset = start_offset
        self._end_offset = end_offset

    def __iter__(self):
        self._fd.seek(self._start_offset)
        # next(f) breaks f.tell(), hence readline() must be used
        line = _safe_readline(self._fd)
        while line:
            pos = self._fd.tell()
            # f.tell() does not always give the byte position in the file
            # sometimes it skips to a very large number
            # it is unlikely that through a normal read we go from
            # end bytes to end + 2**32 bytes (4 GB) and this makes it unlikely
            # that the procedure breaks by the undeterministic behavior of
            # f.tell()
            if (
                self._end_offset > 0
                and pos > self._end_offset
                and pos < self._end_offset + 2**32
            ):
                break
            yield line
            line = self._fd.readline()


class Chunker:
    """
    contextmanager to read a chunck of a file line by line.
    """

    def __init__(self, path: str, start_offset: int, end_offset: int):
        self.path = path
        self.start_offset = start_offset
        self.end_offset = end_offset

    def __enter__(self) -> ChunkLineIterator:
        self.fd = open(self.path, "r", encoding="utf-8")
        return ChunkLineIterator(self.fd, self.start_offset, self.end_offset)

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.fd.close()


def tokenize_line(line):
    line = SPACE_NORMALIZER.sub(" ", line)
    line = line.strip()
    return line.split()


class Dictionary:
    """A mapping from symbols to consecutive integers"""

    def __init__(
        self,
        *,  # begin keyword-only arguments
        bos="<s>",
        pad="<pad>",
        eos="</s>",
        unk="<unk>",
        extra_special_symbols=None,
    ):
        self.bos_word, self.unk_word, self.pad_word, self.eos_word = bos, unk, pad, eos
        self.symbols = []
        self.count = []
        self.indices = {}
        self.bos_index = self.add_symbol(bos)
        self.pad_index = self.add_symbol(pad)
        self.eos_index = self.add_symbol(eos)
        self.unk_index = self.add_symbol(unk)
        if extra_special_symbols:
            for s in extra_special_symbols:
                self.add_symbol(s)
        self.nspecial = len(self.symbols)

    def __eq__(self, other):
        return self.indices == other.indices

    def __getitem__(self, idx):
        if idx < len(self.symbols):
            return self.symbols[idx]
        return self.unk_word

    def get_count(self, idx):
        return self.count[idx]

    def __len__(self):
        """Returns the number of symbols in the dictionary"""
        return len(self.symbols)

    def __contains__(self, sym):
        return sym in self.indices

    def index(self, sym):
        """Returns the index of the specified symbol"""
        assert isinstance(sym, str)
        if sym in self.indices:
            return self.indices[sym]
        return self.unk_index

    def string(
        self,
        tensor,
        bpe_symbol=None,
        escape_unk=False,
        extra_symbols_to_ignore=None,
        unk_string=None,
        include_eos=False,
        separator=" ",
    ):
        """Helper for converting a tensor of token indices to a string.

        Can optionally remove BPE symbols or escape <unk> words.
        """
        if torch.is_tensor(tensor) and tensor.dim() == 2:
            return "\n".join(
                self.string(
                    t,
                    bpe_symbol,
                    escape_unk,
                    extra_symbols_to_ignore,
                    include_eos=include_eos,
                )
                for t in tensor
            )

        extra_symbols_to_ignore = set(extra_symbols_to_ignore or [])
        if not include_eos:
            extra_symbols_to_ignore.add(self.eos())

        def token_string(i):
            if i == self.unk():
                if unk_string is not None:
                    return unk_string
                else:
                    return self.unk_string(escape_unk)
            else:
                return self[i]

        if hasattr(self, "bos_index"):
            extra_symbols_to_ignore.add(self.bos())

        sent = separator.join(
            token_string(i)
            for i in tensor
            if utils_item(i) not in extra_symbols_to_ignore
        )

        return post_process(sent, bpe_symbol)

    def unk_string(self, escape=False):
        """Return unknown string, optionally escaped as: <<unk>>"""
        if escape:
            return "<{}>".format(self.unk_word)
        else:
            return self.unk_word

    def add_symbol(self, word, n=1, overwrite=False):
        """Adds a word to the dictionary"""
        if word in self.indices and not overwrite:
            idx = self.indices[word]
            self.count[idx] = self.count[idx] + n
            return idx
        else:
            idx = len(self.symbols)
            self.indices[word] = idx
            self.symbols.append(word)
            self.count.append(n)
            return idx

    def update(self, new_dict):
        """Updates counts from new dictionary."""
        for word in new_dict.symbols:
            idx2 = new_dict.indices[word]
            if word in self.indices:
                idx = self.indices[word]
                self.count[idx] = self.count[idx] + new_dict.count[idx2]
            else:
                idx = len(self.symbols)
                self.indices[word] = idx
                self.symbols.append(word)
                self.count.append(new_dict.count[idx2])

    def finalize(self, threshold=-1, nwords=-1, padding_factor=8):
        """Sort symbols by frequency in descending order, ignoring special ones.

        Args:
            - threshold defines the minimum word count
            - nwords defines the total number of words in the final dictionary,
                including special symbols
            - padding_factor can be used to pad the dictionary size to be a
                multiple of 8, which is important on some hardware (e.g., Nvidia
                Tensor Cores).
        """
        if nwords <= 0:
            nwords = len(self)

        new_indices = dict(zip(self.symbols[: self.nspecial], range(self.nspecial)))
        new_symbols = self.symbols[: self.nspecial]
        new_count = self.count[: self.nspecial]

        c = Counter(
            dict(
                sorted(zip(self.symbols[self.nspecial :], self.count[self.nspecial :]))
            )
        )
        for symbol, count in c.most_common(nwords - self.nspecial):
            if count >= threshold:
                new_indices[symbol] = len(new_symbols)
                new_symbols.append(symbol)
                new_count.append(count)
            else:
                break

        assert len(new_symbols) == len(new_indices)

        self.count = list(new_count)
        self.symbols = list(new_symbols)
        self.indices = new_indices

        self.pad_to_multiple_(padding_factor)

    def pad_to_multiple_(self, padding_factor):
        """Pad Dictionary size to be a multiple of *padding_factor*."""
        if padding_factor > 1:
            i = 0
            while len(self) % padding_factor != 0:
                symbol = "madeupword{:04d}".format(i)
                self.add_symbol(symbol, n=0)
                i += 1

    def bos(self):
        """Helper to get index of beginning-of-sentence symbol"""
        return self.bos_index

    def pad(self):
        """Helper to get index of pad symbol"""
        return self.pad_index

    def eos(self):
        """Helper to get index of end-of-sentence symbol"""
        return self.eos_index

    def unk(self):
        """Helper to get index of unk symbol"""
        return self.unk_index

    @classmethod
    def load(cls, f):
        """Loads the dictionary from a text file with the format:

        ```
        <symbol0> <count0>
        <symbol1> <count1>
        ...
        ```
        """
        d = cls()
        d.add_from_file(f)
        return d

    def add_from_file(self, f):
        """
        Loads a pre-existing dictionary from a text file and adds its symbols
        to this instance.
        """
        if isinstance(f, str):
            try:
                with open(f, "r", encoding="utf-8") as fd:
                    self.add_from_file(fd)
            except FileNotFoundError as fnfe:
                raise fnfe
            except UnicodeError:
                raise Exception(
                    "Incorrect encoding detected in {}, please "
                    "rebuild the dataset".format(f)
                )
            return

        lines = f.readlines()
        indices_start_line = self._load_meta(lines)

        for line in lines[indices_start_line:]:
            try:
                line, field = line.rstrip().rsplit(" ", 1)
                if field == "#fairseq:overwrite":
                    overwrite = True
                    line, field = line.rsplit(" ", 1)
                else:
                    overwrite = False
                count = int(field)
                word = line
                if word in self and not overwrite:
                    raise RuntimeError(
                        "Duplicate word found when loading Dictionary: '{}'. "
                        "Duplicate words can overwrite earlier ones by adding the "
                        "#fairseq:overwrite flag at the end of the corresponding row "
                        "in the dictionary file. If using the Camembert model, please "
                        "download an updated copy of the model file.".format(word)
                    )
                self.add_symbol(word, n=count, overwrite=overwrite)
            except ValueError:
                raise ValueError(
                    f"Incorrect dictionary format, expected '<token> <cnt> [flags]': \"{line}\""
                )

    def _save(self, f, kv_iterator):
        if isinstance(f, str):
            Path(f).mkdir(exist_ok=True, parents=True)
            with open(f, "w", encoding="utf-8") as fd:
                return self.save(fd)
        for k, v in kv_iterator:
            print("{} {}".format(k, v), file=f)

    def _get_meta(self):
        return [], []

    def _load_meta(self, lines):
        return 0

    def save(self, f):
        """Stores dictionary into a text file"""
        ex_keys, ex_vals = self._get_meta()
        self._save(
            f,
            zip(
                ex_keys + self.symbols[self.nspecial :],
                ex_vals + self.count[self.nspecial :],
            ),
        )

    def dummy_sentence(self, length):
        t = torch.Tensor(length).uniform_(self.nspecial + 1, len(self)).long()
        t[-1] = self.eos()
        return t

    def encode_line(
        self,
        line,
        line_tokenizer=tokenize_line,
        add_if_not_exist=True,
        consumer=None,
        append_eos=True,
        reverse_order=False,
    ) -> torch.IntTensor:
        words = line_tokenizer(line)
        if reverse_order:
            words = list(reversed(words))
        nwords = len(words)
        ids = torch.IntTensor(nwords + 1 if append_eos else nwords)

        for i, word in enumerate(words):
            if add_if_not_exist:
                idx = self.add_symbol(word)
            else:
                idx = self.index(word)
            if consumer is not None:
                consumer(word, idx)
            ids[i] = idx
        if append_eos:
            ids[nwords] = self.eos_index
        return ids

    @staticmethod
    def _add_file_to_dictionary_single_worker(
        filename,
        tokenize,
        eos_word,
        start_offset,
        end_offset,
    ):
        counter = Counter()
        with Chunker(filename, start_offset, end_offset) as line_iterator:
            for line in line_iterator:
                for word in tokenize(line):
                    counter.update([word])
                counter.update([eos_word])
        return counter

    @staticmethod
    def add_file_to_dictionary(filename, dict, tokenize, num_workers):
        def merge_result(counter):
            for w, c in sorted(counter.items()):
                dict.add_symbol(w, c)

        local_file = filename
        offsets = find_offsets(local_file, num_workers)
        if num_workers > 1:
            chunks = zip(offsets, offsets[1:])
            pool = Pool(processes=num_workers)
            results = []
            for start_offset, end_offset in chunks:
                results.append(
                    pool.apply_async(
                        Dictionary._add_file_to_dictionary_single_worker,
                        (
                            local_file,
                            tokenize,
                            dict.eos_word,
                            start_offset,
                            end_offset,
                        ),
                    )
                )
            pool.close()
            pool.join()
            for r in results:
                merge_result(r.get())
        else:
            merge_result(
                Dictionary._add_file_to_dictionary_single_worker(
                    local_file, tokenize, dict.eos_word, offsets[0], offsets[1]
                )
            )



import logging
import re
from pathlib import Path

import torch
import torch.nn as nn
from torch import Tensor


class Lambda(nn.Module):
    """[NOT USED] Custom tensorflow-like Lambda function layer."""

    def __init__(self, function):
        super(Lambda, self).__init__()
        self.function = function

    def forward(self, x: Tensor) -> Tensor:
        return self.function(x)


class NetworkCommonMixIn:
    """Common mixin for network definition."""

    def load_weight(self, weight_file, device):
        """Utility to load a weight file to a device."""

        state_dict = torch.load(weight_file, map_location=device)
        if "state_dict" in state_dict:
            state_dict = state_dict["state_dict"]
        # Remove unneeded prefixes from the keys of parameters.
        weights = {}
        for k in state_dict:
            m = re.search(r"(^fc\.|\.fc\.|^features\.|\.features\.)", k)
            if m is None:
                continue
            new_k = k[m.start() :]
            new_k = new_k[1:] if new_k[0] == "." else new_k
            weights[new_k] = state_dict[k]
        # Load weights and set model to eval().
        self.load_state_dict(weights)
        self.eval()
        logging.info(
            f"Using audio embbeding network pretrained weight: {Path(weight_file).name}"
        )
        return self

    def set_trainable(self, trainable=False):
        for p in self.parameters():
            if p.requires_grad:
                p.requires_grad = trainable



import tempfile
from pathlib import Path

import pytest
import yaml
from dotenv import dotenv_values
from torch.utils.data import Subset
from tqdm import tqdm

from s3prl.downstream.emotion.expert import DownstreamExpert
from s3prl.problem import SuperbER


@pytest.mark.corpus
@pytest.mark.parametrize("fold_id", [0, 1, 2, 3, 4])
def test_er_dataset(fold_id):
    v3_er_folder = Path(__file__).parent.parent / "s3prl" / "downstream" / "emotion"

    IEMOCAP = dotenv_values()["IEMOCAP"]
    with (v3_er_folder / "config.yaml").open() as file:
        config = yaml.load(file, Loader=yaml.FullLoader)["downstream_expert"]
        config["datarc"]["root"] = IEMOCAP
        config["datarc"]["meta_data"] = v3_er_folder / "meta_data"
        config["datarc"]["test_fold"] = f"fold{fold_id + 1}"

    with tempfile.TemporaryDirectory() as tempdir:
        expert = DownstreamExpert(320, config, tempdir)

        train_dataset_v3 = expert.get_dataloader("train").dataset
        valid_dataset_v3 = expert.get_dataloader("dev").dataset
        test_dataset_v3 = expert.get_dataloader("test").dataset

    with tempfile.TemporaryDirectory() as tempdir:
        default_config = SuperbER().default_config()
        train_csv, valid_csv, test_csvs = SuperbER().prepare_data(
            {"iemocap": IEMOCAP, "test_fold": fold_id}, tempdir, tempdir
        )
        encoder_path = SuperbER().build_encoder(
            default_config["build_encoder"],
            tempdir,
            tempdir,
            train_csv,
            valid_csv,
            test_csvs,
        )
        train_dataset_v4 = SuperbER().build_dataset(
            default_config["build_dataset"],
            tempdir,
            tempdir,
            "train",
            train_csv,
            encoder_path,
            None,
        )
        valid_dataset_v4 = SuperbER().build_dataset(
            default_config["build_dataset"],
            tempdir,
            tempdir,
            "valid",
            valid_csv,
            encoder_path,
            None,
        )
        test_dataset_v4 = SuperbER().build_dataset(
            default_config["build_dataset"],
            tempdir,
            tempdir,
            "test",
            test_csvs[0],
            encoder_path,
            None,
        )

    def compare_dataset(v3, v4):
        data_v3 = {}
        for wav, label, name in tqdm(v3, desc="v3"):
            if isinstance(v3, Subset):
                v3 = v3.dataset
            label_name = [k for k, v in v3.class_dict.items() if v == label][0]
            data_v3[name] = label_name

        data_v4 = {}
        for batch in tqdm(v4, desc="v4"):
            data_v4[batch["unique_name"]] = batch["label"]

        assert sorted(data_v3.keys()) == sorted(data_v4.keys())
        for key in data_v3:
            value_v3 = data_v3[key]
            value_v4 = data_v4[key]
            assert value_v3 == value_v4

    compare_dataset(train_dataset_v3, train_dataset_v4)
    compare_dataset(valid_dataset_v3, valid_dataset_v4)
    compare_dataset(test_dataset_v3, test_dataset_v4)



import json
import pickle
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
import torch
from omegaconf import MISSING

from s3prl.dataio.dataset import EncodeMultiLabel, LoadAudio
from s3prl.dataio.encoder import CategoryEncoder
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler
from s3prl.nn.hear import HearFullyConnectedPrediction
from s3prl.task.scene_prediction import ScenePredictionTask

from ._hear_util import resample_hear_corpus
from .superb_sid import SuperbSID

__all__ = ["HearFSD"]


def hear_scene_trainvaltest(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    get_path_only: bool = False,
):
    target_dir = Path(target_dir)

    resample_hear_corpus(dataset_root, target_sr=16000)

    dataset_root = Path(dataset_root)
    wav_root: Path = dataset_root / "16000"

    train_csv = target_dir / "train.csv"
    valid_csv = target_dir / "valid.csv"
    test_csv = target_dir / "test_csv"

    if get_path_only:
        return train_csv, valid_csv, [test_csv]

    def load_json(filepath):
        with open(filepath, "r") as fp:
            return json.load(fp)

    def split_to_df(split: str) -> pd.DataFrame:
        meta = load_json(dataset_root / f"{split}.json")
        data = defaultdict(list)
        for k in list(meta.keys()):
            data["id"].append(k)
            data["wav_path"].append(wav_root / split / k)
            data["labels"].append(" ; ".join([str(label).strip() for label in meta[k]]))
        return pd.DataFrame(data=data)

    split_to_df("train").to_csv(train_csv, index=False)
    split_to_df("valid").to_csv(valid_csv, index=False)
    split_to_df("test").to_csv(test_csv, index=False)

    return train_csv, valid_csv, [test_csv]


class HearFSD(SuperbSID):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=10,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_layers=2,
                pooling_type="MeanPooling",
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                prediction_type="multilabel",
                scores=["mAP", "top1_acc", "d_prime", "aucroc"],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-3,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=40000,
                log_step=100,
                eval_step=1000,
                save_step=100,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="mAP",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        return hear_scene_trainvaltest(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_encoder(
        self,
        build_encoder: dict,
        target_dir: str,
        cache_dir: str,
        train_csv_path: str,
        valid_csv_path: str,
        test_csv_paths: list,
        get_path_only: bool = False,
    ):
        encoder_path = Path(target_dir) / "encoder.pkl"
        if get_path_only:
            return encoder_path

        train_csv = pd.read_csv(train_csv_path)
        valid_csv = pd.read_csv(valid_csv_path)
        test_csvs = [pd.read_csv(path) for path in test_csv_paths]
        all_csv = pd.concat([train_csv, valid_csv, *test_csvs])
        all_labels = []
        for rowid, row in all_csv.iterrows():
            labels = str(row["labels"]).split(";")
            labels = [l.strip() for l in labels]
            all_labels.extend(labels)

        encoder = CategoryEncoder(all_labels)
        with open(encoder_path, "wb") as f:
            pickle.dump(encoder, f)

        return encoder_path

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
        frame_shift: int,
    ):
        df = pd.read_csv(data_csv)
        ids = df["id"].tolist()
        wav_paths = df["wav_path"].tolist()
        labels = [
            [single_label.strip() for single_label in str(label_str).split(";")]
            for label_str in df["labels"].tolist()
        ]
        with open(encoder_path, "rb") as f:
            encoder = pickle.load(f)

        audio_loader = LoadAudio(wav_paths)
        label_encoder = EncodeMultiLabel(labels, encoder)

        class Dataset:
            def __len__(self):
                return len(audio_loader)

            def __getitem__(self, index: int):
                audio = audio_loader[index]
                label = label_encoder[index]
                return {
                    "x": audio["wav"],
                    "x_len": audio["wav_len"],
                    "y": label["binary_labels"],
                    "labels": label["labels"],
                    "unique_name": ids[index],
                }

        dataset = Dataset()
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_batch_sampler)
        return FixedBatchSizeBatchSampler(dataset, **(conf.train or {}))

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        return HearFullyConnectedPrediction(
            downstream_input_size, downstream_output_size, **build_downstream
        )

    def build_task(
        self,
        build_task: dict,
        model: torch.nn.Module,
        encoder,
        valid_df: pd.DataFrame = None,
        test_df: pd.DataFrame = None,
    ):
        return ScenePredictionTask(model, encoder, **build_task)



"""
RNN models used in Superb Benchmark

Authors:
  * Heng-Jui Chang 2022
  * Leo 2022
"""

from typing import List, Tuple

import torch
import torch.nn as nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

from s3prl.nn.interface import AbsFrameModel

__all__ = ["RNNEncoder", "SuperbDiarizationModel", "RNNLayer"]


def downsample(
    x: torch.Tensor, x_len: torch.LongTensor, sample_rate: int, sample_style: str
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Downsamples a sequence.

    Args:
        x (torch.Tensor): Sequence (batch, timestamps, hidden_size)
        x_len (torch.LongTensor): Sequence length (batch, )
        sample_rate (int): Downsample rate (must be greater than one)
        sample_style (str): Downsample style ("drop" or "concat")

    Raises:
        NotImplementedError: Sample style not supported.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]:
            x (torch.Tensor): (batch, timestamps // sample_rate, output_size)
            x_len (torch.LongTensor): (batch, )
    """

    B, T, D = x.shape
    x_len = torch.div(x_len, sample_rate, rounding_mode="floor")

    if sample_style == "drop":
        # Drop the unselected timesteps
        x = x[:, ::sample_rate, :].contiguous()
    elif sample_style == "concat":
        # Drop the redundant frames and concat the rest according to sample rate
        if T % sample_rate != 0:
            x = x[:, : -(T % sample_rate), :]
        x = x.contiguous().view(B, int(T / sample_rate), D * sample_rate)
    else:
        raise NotImplementedError(f"Sample style={sample_style} not supported.")

    return x, x_len


class RNNLayer(nn.Module):
    """RNN Layer

    Args:
        input_size (int): Input size.
        hidden_size (int): Hidden size.
        module (str): RNN module (RNN, GRU, LSTM)
        dropout (float, optional): Dropout rate. Defaults to 0.0.
        bidirectional (bool, optional): Bidirectional. Defaults to False.
        proj (bool, optional): Projection layer. Defaults to False.
        layer_norm (bool, optional): Layer normalization. Defaults to False.
        sample_rate (int, optional): Downsampling rate. Defaults to 1.
        sample_style (str, optional): Downsampling style (**drop** or **concat**). Defaults to "drop".
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        module: str,
        dropout: float = 0.0,
        bidirectional: bool = False,
        proj: bool = False,
        layer_norm: bool = False,
        sample_rate: int = 1,
        sample_style: str = "drop",
    ):
        super().__init__()
        self._insize = input_size

        self.out_size = (
            hidden_size
            * (2 if bidirectional else 1)
            * (2 if sample_style == "concat" and sample_rate > 1 else 1)
        )
        self.dropout = dropout
        self.proj = proj
        self.layer_norm = layer_norm
        self.sample_rate = sample_rate
        self.sample_style = sample_style

        assert module.upper() in {"RNN", "GRU", "LSTM"}
        assert sample_style in {"drop", "concat"}

        self.layer = getattr(nn, module.upper())(
            input_size,
            hidden_size,
            num_layers=1,
            batch_first=True,
            bidirectional=bidirectional,
        )

        if self.layer_norm:
            rnn_out_size = hidden_size * (2 if bidirectional else 1)
            self.ln_layer = nn.LayerNorm(rnn_out_size)

        if self.dropout > 0:
            self.dp_layer = nn.Dropout(self.dropout)

        if self.proj:
            self.pj_layer = nn.Linear(self.out_size, self.out_size)

    def forward(self, xs: torch.Tensor, xs_len: torch.LongTensor):
        """
        Args:
            xs (torch.FloatTensor): (batch_size, seq_len, input_size)
            xs_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple:

            1. ys (torch.FloatTensor): (batch_size, seq_len, output_size)
            2. ys_len (torch.LongTensor): (batch_size, )
        """
        if not self.training:
            self.layer.flatten_parameters()

        xs = pack_padded_sequence(
            xs, xs_len.cpu(), batch_first=True, enforce_sorted=False
        )
        output, _ = self.layer(xs)
        output, _ = pad_packed_sequence(output, batch_first=True)

        # Normalization
        if self.layer_norm:
            output = self.ln_layer(output)

        if self.dropout > 0:
            output = self.dp_layer(output)

        # Downsampling
        if self.sample_rate > 1:
            output, xs_len = downsample(
                output, xs_len, self.sample_rate, self.sample_style
            )

        # Projection
        if self.proj:
            output = torch.tanh(self.pj_layer(output))

        return output, xs_len

    @property
    def input_size(self) -> int:
        return self._insize

    @property
    def output_size(self) -> int:
        return self.out_size


class RNNEncoder(AbsFrameModel):
    """RNN Encoder for sequence to sequence modeling, e.g., ASR.

    Args:
        input_size (int): Input size.
        output_size (int): Output size.
        module (str, optional): RNN module type. Defaults to "LSTM".
        hidden_size (List[int], optional): Hidden sizes for each layer. Defaults to [1024].
        dropout (List[float], optional): Dropout rates for each layer. Defaults to [0.0].
        layer_norm (List[bool], optional): Whether to use layer norm for each layer. Defaults to [False].
        proj (List[bool], optional): Whether to use projection for each layer. Defaults to [True].
        sample_rate (List[int], optional): Downsample rates for each layer. Defaults to [1].
        sample_style (str, optional): Downsample style ("drop" or "concat"). Defaults to "drop".
        bidirectional (bool, optional): Whether RNN layers are bidirectional. Defaults to False.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        module: str = "LSTM",
        proj_size: int = 1024,
        hidden_size: List[int] = [1024],
        dropout: List[float] = [0.0],
        layer_norm: List[bool] = [False],
        proj: List[bool] = [True],
        sample_rate: List[int] = [1],
        sample_style: str = "drop",
        bidirectional: bool = False,
    ):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

        prev_size = input_size

        self.proj = nn.Linear(prev_size, proj_size)
        prev_size = proj_size

        self.rnns = nn.ModuleList()
        for i in range(len(hidden_size)):
            rnn_layer = RNNLayer(
                input_size=prev_size,
                hidden_size=hidden_size[i],
                module=module,
                dropout=dropout[i],
                bidirectional=bidirectional,
                proj=proj[i],
                layer_norm=layer_norm[i],
                sample_rate=sample_rate[i],
                sample_style=sample_style,
            )
            self.rnns.append(rnn_layer)
            prev_size = rnn_layer.output_size

        self.linear = nn.Linear(prev_size, output_size)

    def forward(self, x: torch.Tensor, x_len: torch.LongTensor):
        """
        Args:
            xs (torch.FloatTensor): (batch_size, seq_len, input_size)
            xs_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple:

            1. ys (torch.FloatTensor): (batch_size, seq_len, output_size)
            2. ys_len (torch.LongTensor): (batch_size, )
        """

        xs, xs_len = x, x_len
        xs = self.proj(xs)

        for rnn in self.rnns:
            xs, xs_len = rnn(xs, xs_len)

        logits = self.linear(xs)

        return logits, xs_len

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def output_size(self) -> int:
        return self._output_size


class SuperbDiarizationModel(AbsFrameModel):
    """
    The exact RNN model used in SUPERB Benchmark for Speaker Diarization

    Args:
        input_size (int): input_size
        output_size (int): output_size
        rnn_layers (int): number of rnn layers
        hidden_size (int): the hidden size across all rnn layers
    """

    def __init__(
        self, input_size: int, output_size: int, rnn_layers: int, hidden_size: int
    ):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

        self.use_rnn = rnn_layers > 0
        if self.use_rnn:
            self.rnn = nn.LSTM(
                input_size, hidden_size, num_layers=rnn_layers, batch_first=True
            )
            self.linear = nn.Linear(hidden_size, output_size)
        else:
            self.linear = nn.Linear(input_size, output_size)

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def output_size(self) -> int:
        return self._output_size

    def forward(self, xs, xs_len):
        """
        Args:
            xs (torch.FloatTensor): (batch_size, seq_len, input_size)
            xs_len (torch.LongTensor): (batch_size, )

        Returns:
            tuple:

            1. ys (torch.FloatTensor): (batch_size, seq_len, output_size)
            2. ys_len (torch.LongTensor): (batch_size, )
        """
        features, features_len = xs, xs_len
        features = features.float()
        if self.use_rnn:
            hidden, _ = self.rnn(features)
            predicted = self.linear(hidden)
        else:
            predicted = self.linear(features)

        return predicted, features_len



# -*- coding: utf-8 -*-
# @Time    : 8/25/21 5:25 PM
# @Author  : Yuan Gong
# @Affiliation  : Massachusetts Institute of Technology
# @Email   : yuangong@mit.edu
# @File    : expert.py

# Author
# - Leo

import logging

import torch

from ..ssast.audio import FeatureExtractor

logger = logging.getLogger(__name__)

FBANK_SAMPLE_STRIDE = 160
PATCH_FBANK_STRIDE = 10
SAMPLE_RATE = 16000


class UpstreamExpert(torch.nn.Module):
    def __init__(
        self,
        ckpt: str,
        window_secs: float = 10.24,
        stride_secs: float = 10.24,
        feature_selection: str = "cls",
    ):
        super().__init__()
        assert feature_selection in ["cls", "hidden_states"]
        self.feature_selection = feature_selection

        self.window_secs = window_secs
        self.stride_secs = stride_secs
        target_length = int(window_secs * SAMPLE_RATE / FBANK_SAMPLE_STRIDE)

        try:
            import timm

            from .ast_models import ASTModel
        except:
            logger.error(
                "SSAST requires 'timm==0.4.5' to work. Please run 'pip install timm==0.4.5'"
            )
            exit(1)

        self.preprocessor = FeatureExtractor(
            target_length=target_length, apply_cmvn=False
        )
        self.model = ASTModel(
            label_dim=527,
            fstride=10,
            tstride=PATCH_FBANK_STRIDE,
            input_fdim=128,
            input_tdim=int(window_secs * 100),
            imagenet_pretrain=True,
            audioset_pretrain=True,
            model_size="base384",
            pretrained_ckpt=ckpt,
        ).cpu()  # ensure the entire model is on cpu

    def get_downsample_rates(self, key: str = None) -> int:
        if self.feature_selection == "cls":
            return int(self.stride_secs * SAMPLE_RATE)
        elif self.feature_selection == "hidden_states":
            return int(FBANK_SAMPLE_STRIDE * PATCH_FBANK_STRIDE)

    def forward(self, wavs):
        wavs_len = [len(wav) for wav in wavs]
        max_wav_len = max(wavs_len)
        start_points = list(range(0, max_wav_len, int(self.stride_secs * SAMPLE_RATE)))
        padded_max_wav_len = start_points[-1] + int(self.window_secs * SAMPLE_RATE)
        padded_wavs = [
            torch.cat([wav, wav.new_zeros(padded_max_wav_len - len(wav))])
            for wav in wavs
        ]

        all_features = []
        for start in start_points:
            subwavs = [
                wav[start : start + int(self.window_secs * SAMPLE_RATE)]
                for wav in padded_wavs
            ]
            features = [self.preprocessor(wav.unsqueeze(0)) for wav in subwavs]
            features = torch.stack(
                features, dim=0
            )  # (batch_size, segment_seq_len, hidden_size)
            all_features.append(features)

        all_features = torch.stack(all_features, dim=0)
        num_segment, batch_size, segment_seq_len, hidden_size = all_features.shape

        flatten_features = all_features.reshape(-1, segment_seq_len, hidden_size)
        output, hidden_states = self.model(flatten_features)
        # output: (num_segment * batch_size, num_class)
        # hidden_states: List[(num_segment * batch_size, segment_seq_len, hidden_size)]

        if self.feature_selection == "cls":
            output = output.reshape(num_segment, batch_size, -1).transpose(0, 1).float()
            # (batch_size, num_segment, num_class)
            hidden_states = [output]

        elif self.feature_selection == "hidden_states":
            reshaped_hidden_states = [
                (
                    h.reshape(num_segment, batch_size, -1, h.size(-1))
                    .transpose(
                        0, 1
                    )  # (batch_size, num_segment, num_horizon_patch, num_vertical_patch * hidden_size)
                    .flatten(
                        1, 2
                    )  # (batch_size, num_segment * num_horizon_patch, num_vertical_patch * hidden_size)
                    .float()
                )
                for h in hidden_states
            ]
            hidden_states = reshaped_hidden_states

        trimmed_hidden_states = []
        for h in hidden_states:
            max_h_len = len(range(0, max_wav_len, self.get_downsample_rates()))
            h = h[:, :max_h_len, :]
            trimmed_hidden_states.append(h)

        return {
            "hidden_states": trimmed_hidden_states,
        }



"""
The shared backbone of common ML train/test procedure for all problems

Authors:
  * Leo 2022
"""

from __future__ import annotations

import argparse
import inspect
import logging
import math
import os
import pickle
import shutil
import sys
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from time import time
from typing import Dict, List, Union

import omegaconf
import torch
import yaml
from torch.utils.data import DataLoader
from torch.utils.tensorboard.writer import SummaryWriter
from tqdm import tqdm

from s3prl.dataio.collate_fn import default_collate_fn
from s3prl.dataio.sampler import DistributedBatchSamplerWrapper
from s3prl.nn.upstream import Featurizer, S3PRLUpstream, UpstreamDownstreamModel
from s3prl.task import Task
from s3prl.util.override import parse_overrides
from s3prl.util.seed import fix_random_seeds

logger = logging.getLogger(__name__)

LOGGING_FORMAT = "%(levelname)s | %(asctime)s | %(module)s:%(lineno)d | %(message)s"
ACCEPTABLE_ERRORS = [
    "CUDA out of memory",
    "Unable to find a valid cuDNN algorithm to run convolution",  # Usually caused by CUDA OOM
]
PRIMITIVE_TYPES = (int, float, bool, str)

DEFAULT_CONFIG_FORMAT = """
The default arguments for :obj:`run` in yaml.
Note that for the fields with inner values, like :code:`build_model`,
the outer field name corresponds to a method name, so you can find the method
:obj:`build_model`. Furthermore, the values inside that field will be
directly passed into the method. So by changing these inner values, you
can directly affect the behavior of the corresponding method. See the method
documentation for all the supported arguments and their meanings.

The methods affected by the following config are: {:s}

.. code-block:: yaml

{:s}
"""

__all__ = ["Problem"]


class _DistributedDataParallel(torch.nn.parallel.DistributedDataParallel):
    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)


def _force_cacheable(data: dict):
    output = dict()
    for key, value in data.items():
        if isinstance(value, torch.Tensor):
            value = value.detach().cpu()
        output[key] = value
    return output


def _to_device(data, device: str):
    output = dict()
    for key, value in data.items():
        if isinstance(value, torch.Tensor):
            value = value.to(device)
        output[key] = value
    return output


def _doc_default_config(cls: Problem):
    """
    This is used to layout the :code:`default_config` dictionary into yaml format
    for :code:`default_config`'s docstring.
    """

    def _append_prefix_spaces(docstring: str):
        return "\n".join([f"  {line}" for line in docstring.split("\n")])

    obj = cls()
    try:
        config = obj.default_config()
    except:
        return
    else:
        methods = []
        for k, v in config.items():
            if hasattr(cls, k):
                methods.append(getattr(cls, k))
        method_links = " ".join([f":obj:`{method.__name__}`" for method in methods])

        yaml_str = yaml.dump(config, sort_keys=False, width=float("inf"))
        yaml_str = _append_prefix_spaces(yaml_str)
        cls.default_config.__doc__ = DEFAULT_CONFIG_FORMAT.format(
            method_links, yaml_str
        )


class Problem:
    _store: Dict[str, Problem] = dict()

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls._store[cls.__name__] = cls
        _doc_default_config(cls)

    @classmethod
    def get_class_from_name(cls, name: str):
        """
        Args:
            name (str): the :code:`__name__` of the problem class

        Returns:
            Problem
        """
        assert (
            name in cls._store
        ), f"The class '{name}' is either not defined or not imported"
        return cls._store[name]

    def build_collate_fn(self, build_collate_fn: dict, mode: str):
        """
        By default returns :obj:`s3prl.dataset.base.default_collate_fn`

        Args:
            build_collate_fn (dict): same in :obj:`default_config`, no argument supported for now
            mode (str): train, valid, or test

        Returns:
            callable

            the collate_fn for torch DataLoader in train/valid/test :code:`mode`
        """
        return default_collate_fn

    def build_upstream(self, build_upstream: dict):
        """
        By default build the upstream with :obj:`s3prl.nn.upstream.S3PRLUpstream`

        Args:
            build_upstream (dict): same in :obj:`default_config`,
                arguments for :obj:`s3prl.nn.upstream.S3PRLUpstream`

        Returns:
            :obj:`s3prl.nn.interface.AbsUpstream`

            Return an upstream model, whose forward takes the waveform input and returns
            multiple hidden states as features.
        """
        upstream = S3PRLUpstream(**build_upstream)
        return upstream

    def build_featurizer(self, build_featurizer: dict, upstream):
        """
        By default build the featurizer with :obj:`s3prl.nn.Featurizer`

        Args:
            build_featurizer (dict): same in :obj:`default_config`,
                arguments for :obj:`s3prl.nn.Featurizer`
            upstream (:obj:`AbsUpstream`): the upstream model built by :obj:`build_upstream`

        Returns:
            :obj:`s3prl.nn.interface.AbsFeaturizer`

            Return the featurizer model. The featurizer is used to reduce the multiple
            hidden states returned from the upstream model (built by :obj:`build_upstream`)
            into a single hidden state, so can be easliy fed into the downstream model
        """
        featurizer = Featurizer(upstream, **build_featurizer)
        return featurizer

    def build_model(
        self,
        build_model: dict,
        model_output_size: int,
        build_upstream: dict,
        build_featurizer: dict,
        build_downstream: dict,
    ):
        """
        By default build model with :obj:`s3prl.nn.upstream.UpstreamDownstreamModel`

        Args:
            build_model (dict): same in :obj:`default_config`,
                arguments for :obj:`s3prl.nn.upstream.UpstreamDownstreamModel`
            model_output_size (int): the required model's output hidden size
            build_upstream (dict): same in :obj:`default_config`, refer to :obj:`build_upstream`
            build_featurizer (dict): same in :obj:`default_config`, refer to :obj:`build_featurizer`
            build_downstream (dict): same in :obj:`default_config`, refer to :obj:`build_downstream`

        Returns:
            torch.nn.Module

            Return the entire model for the task, which takes the direct items from DataLoader as the input.
            Usually, the components can be built by :obj:`build_upstream`, :obj:`build_featurizer`,
            :obj:`build_downstream`, and are concated together to get the final model.
            The upstream extracts multiple hidden states, the featuizer reduce them into a single hidden state,
            and the downstream takes the hidden states as the feature for the downstream-specific model.
        """
        upstream = self.build_upstream(build_upstream)
        featurizer: Featurizer = self.build_featurizer(build_featurizer, upstream)
        downstream = self.build_downstream(
            build_downstream,
            featurizer.output_size,
            model_output_size,
            featurizer.downsample_rate,
        )
        model = UpstreamDownstreamModel(upstream, featurizer, downstream, **build_model)
        return model

    def build_optimizer(self, build_optimizer: dict, parameters):
        """
        Args:
            build_optimizer (dict): same in :obj:`default_config`, refer to below

                ====================  ====================
                key                   description
                ====================  ====================
                name                  (str) - the optimizer class name in :obj:`torch.optim`
                conf                  (dict) - the arguments for initializing the optimizer class. e.g. :code:`{"lr": 1.0e-4}`
                ====================  ====================

            parameters (iterable): the standard params accepted by :obj:`torch.optim.Optimizer`.

        Returns:
            :obj:`torch.optim.Optimizer`

            An optimizer following standard torch usage
        """

        def _default_build_optimizer(name: str, conf: dict):
            opt_cls = getattr(torch.optim, name)
            opt = opt_cls(parameters, **conf)
            return opt

        return _default_build_optimizer(**build_optimizer)

    def build_scheduler(self, build_scheduler: dict, optimizer):
        """
        Args:
            build_scheduler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                name                  (str) - the scheduler class name in :obj:`torch.optim.lr_scheduler`
                conf                  (dict) - the arguments for initializing the scheduler class. e.g. :code:`{"gamma": 0.01}` for :obj:`torch.optim.lr_scheduler.StepLR`
                ====================  ====================

            optimizer: the standard torch optimizer accepted by Scheduler in :obj:`torch.optim.lr_scheduler`.

        Returns:
            torch scheduler

            A scheduler following standard torch usage
        """

        def _default_build_scheduler(name: str, conf: dict):
            scheduler_cls = getattr(torch.optim.lr_scheduler, name)
            scheduler = scheduler_cls(optimizer, **conf)
            return scheduler

        return _default_build_scheduler(**build_scheduler)

    def train(
        self,
        train: dict,
        train_dir: str,
        build_model_all_args: dict,
        build_task_all_args_except_model: dict,
        save_model: dict,
        save_task: dict,
        build_optimizer: dict,
        build_scheduler: dict,
        evaluate: dict,
        train_dataset,
        train_batch_sampler,
        train_collate_fn,
        valid_dataset,
        valid_batch_sampler,
        valid_collate_fn,
        num_workers: int,
        world_size: int,
        rank: int,
        eval_batch: int,
        device: str,
        global_config: dict = None,
    ):
        """
        Args:
            train (dict): same in :obj:`default_config`

                ==========================  ====================
                key                         description
                ==========================  ====================
                total_steps                 (int) - the total optimization steps
                log_step                    (int) - logging frequency. log every :code:`log_step` step
                eval_step                   (int) - evaluation frequency. Evaluate every :code:`eval_step` step. \
                                                Note that you can control how many batch to evaluate to speed up the \
                                                development by the :code:`eval_batch` argument in :obj:`run`
                save_step                   (int) - save the checkpoint every :code:`save_step` step.
                gradient_clipping           (float) - clip the gradient. important for RNNs.
                gradient_accumulate         (int) - accumulate multiple steps' gradient before updating network parameters \
                                                to simulate large-batch optimization.
                valid_metric                (str) - the metric to select the best valid checkpoint. Different Tasks have different \
                                                supported valid_metrics. See :obj:`build_task` for the supported metrics.
                valid_higher_better         (bool) - some metrics are higher better, while some are lower better \
                                                this will affect how to save the best validation checkpoint.
                auto_resume                 (bool) - if there are already the last checkpoint in :code:`target_dir` (see :obj:`run`), \
                                                whether to resume from it or delete it and start a new training session.
                resume_ckpt_dir             (str) - you can directly specify the checkpoint path to resume which is not necessary \
                                                in :code:`target_dir` (see :obj:`run`).
                seed                        (int) - fix the seed before the training start
                keep_num_ckpts              (int) - to prevent saving too many checkpoints, only save the :code:`keep_num_ckpts` \
                                                latest checkpoints and delete the old ones.
                use_scheduler               (bool) - whether to use the scheduler
                ==========================  ====================

            **others:
                only meaningful when you want to override this train method, which is not the
                common case. Hence we skip the documentation for now.
        """

        @dataclass
        class TrainConfig:
            total_steps: int
            log_step: int
            eval_step: int
            save_step: int
            gradient_clipping: float
            gradient_accumulate: int
            valid_metric: str
            valid_higher_better: bool
            auto_resume: bool = True
            resume_ckpt_dir: str = None
            seed: int = 0
            keep_num_ckpts: int = 2
            use_scheduler: bool = False

        conf = TrainConfig(**train)

        fix_random_seeds(conf.seed)

        train_dir: Path = Path(train_dir)
        if not conf.auto_resume and train_dir.is_dir():
            logger.warning(
                f"{train_dir} exists. Delete the directory since auto_resume=False"
            )
            shutil.rmtree(train_dir)
        train_dir.mkdir(exist_ok=True, parents=True)

        ckpt_dirs = [key for key in os.listdir(train_dir) if key.startswith("step_")]
        ckpt_dirs.sort(key=lambda name: int(name.split("_")[-1]), reverse=True)

        resume = False
        if conf.auto_resume:
            if conf.resume_ckpt_dir is not None and Path(conf.resume_ckpt_dir).is_dir():
                resume = True
            if len(ckpt_dirs) > 0:
                resume = True

        if resume:
            resume_ckpt_dir = Path(conf.resume_ckpt_dir or train_dir / ckpt_dirs[0])
            logger.info(f"Loading checkpoints from {resume_ckpt_dir}")
            try:
                _, task = self.load_model_and_task(resume_ckpt_dir)
            except:
                logger.error(
                    f"Fail to load the checkpoint {resume_ckpt_dir}. "
                    "You can set '--train.auto_resume False' to ignore the crashed checkpoint to avoid this behavior."
                )
                raise

            optimizer_state = torch.load(
                resume_ckpt_dir / "optimizer.pt", map_location="cpu"
            )

            if conf.use_scheduler:
                scheduler_state = torch.load(
                    resume_ckpt_dir / "scheduler.pt", map_location="cpu"
                )
            else:
                scheduler_state = None

            with open(resume_ckpt_dir / "training_stats.yaml", "r") as f:
                training_stats = yaml.load(f, Loader=yaml.FullLoader)

            global_step = int(training_stats["global_step"])
            epoch = int(training_stats["epoch"])
            valid_best_metrics = dict(training_stats["valid_best_metrics"])

        else:
            model = self.build_model(**build_model_all_args)
            task = self.build_task(model=model, **build_task_all_args_except_model)
            optimizer_state = None
            scheduler_state = None
            global_step = 0
            epoch = 0
            valid_best_metrics = dict()

        device = torch.device(device)
        wrapped_task = task.to(device)

        if world_size > 1:
            torch.cuda.set_device(device.index)
            wrapped_task = _DistributedDataParallel(
                task,
                device_ids=[device.index],
                find_unused_parameters=True,
                output_device=device.index,
            )

        optimizer = self.build_optimizer(build_optimizer, task.parameters())
        if optimizer_state:
            optimizer.load_state_dict(optimizer_state)

        scheduler = None
        if conf.use_scheduler:
            scheduler = self.build_scheduler(build_scheduler, optimizer)
            if scheduler_state:
                scheduler.load_state_dict(scheduler_state)

        train_batch_sampler = DistributedBatchSamplerWrapper(
            train_batch_sampler,
            num_replicas=world_size,
            rank=rank,
        )
        train_dataloader = DataLoader(
            train_dataset,
            batch_sampler=train_batch_sampler,
            num_workers=num_workers,
            collate_fn=train_collate_fn,
        )

        tqdm_file = sys.stderr if rank == 0 else open(os.devnull, "w")
        pbar = tqdm(
            total=conf.total_steps,
            dynamic_ncols=True,
            desc="train",
            file=tqdm_file,
        )
        pbar.n = global_step

        if rank == 0:
            tf_dir = train_dir / "tb"
            tf_logger = SummaryWriter(str(tf_dir))

        def _save_ckpts_to_dir(
            ckpts_dir: str,
            task,
            optimizer,
            scheduler,
            build_model_all_args: dict,
            build_task_all_args_except_model: dict,
            save_model: dict,
            save_task: dict,
            training_stats: dict,
            global_config: dict,
        ):
            ckpts_dir: Path = Path(ckpts_dir)
            ckpts_dir.mkdir(exist_ok=True, parents=True)

            model_ckpt_dir = ckpts_dir / "model"
            self.save_model(
                save_model, model_ckpt_dir, build_model_all_args, task.model
            )

            task_ckpt_dir = ckpts_dir / "task"
            self.save_task(
                save_task, task_ckpt_dir, build_task_all_args_except_model, task
            )

            torch.save(optimizer.state_dict(), ckpts_dir / "optimizer.pt")
            if scheduler is not None:
                torch.save(scheduler.state_dict(), ckpts_dir / "scheduler.pt")

            with (ckpts_dir / "training_stats.yaml").open("w") as f:
                yaml.safe_dump(training_stats, f)

            with (ckpts_dir / "config.yaml").open("w") as f:
                yaml.safe_dump(global_config, f)

        backward_steps = 0
        while pbar.n < pbar.total:
            train_batch_sampler.set_epoch(epoch),
            batch_results = []
            logger.info(f"Start epoch {epoch}")
            for batch in train_dataloader:
                # try/except block for forward/backward
                try:
                    if pbar.n >= pbar.total:
                        break
                    global_step = pbar.n + 1

                    wrapped_task.train()
                    batch = _to_device(batch, device)
                    loss, cacheable = wrapped_task("train", **batch)
                    (loss / conf.gradient_accumulate).backward()
                    batch_results.append(_force_cacheable(cacheable))

                except RuntimeError as e:
                    if world_size > 1:
                        raise

                    acceptable = False
                    for acc_err in ACCEPTABLE_ERRORS:
                        if str(e) in acc_err:
                            acceptable = True
                            break
                    if not acceptable:
                        raise

                    logger.warning(f"Step {global_step}: {str(e)}")
                    with torch.cuda.device(device):
                        torch.cuda.empty_cache()
                    optimizer.zero_grad()
                    continue

                backward_steps += 1
                if backward_steps % conf.gradient_accumulate > 0:
                    continue

                grad_norm = torch.nn.utils.clip_grad_norm_(
                    wrapped_task.parameters(), conf.gradient_clipping
                )

                if math.isnan(grad_norm):
                    logger.warning(f"[Runner] - grad norm is NaN at step {global_step}")
                else:
                    optimizer.step()
                optimizer.zero_grad()

                if conf.use_scheduler:
                    scheduler.step()

                if rank > 0:
                    batch_results = []
                    pbar.update(1)
                    continue

                def _log_results(
                    split_name: str,
                    logs: dict,
                    tensorboard: SummaryWriter,
                    global_step: int,
                ):
                    logger.info(f"{split_name} at step {global_step}")
                    for name, value in logs.items():
                        value = float(value)
                        logger.info(f"{name}: {value}")
                        tensorboard.add_scalar(
                            f"{split_name}-{name}", value, global_step=global_step
                        )

                if global_step % conf.log_step == 0:
                    logs = wrapped_task.reduction("train", batch_results)
                    _log_results("train", logs, tf_logger, global_step)
                    batch_results = []

                save_names = []

                if global_step % conf.eval_step == 0:
                    assert (
                        valid_dataset is not None and valid_batch_sampler is not None
                    ), f"valid dataset is not supported, please set train.eval_step to infinite"
                    logs: dict = self.evaluate(
                        evaluate,
                        "valid",
                        task,
                        valid_dataset,
                        valid_batch_sampler,
                        valid_collate_fn,
                        eval_batch,
                        train_dir,
                        device,
                        num_workers,
                    )
                    _log_results("valid", logs, tf_logger, global_step)
                    valid_metrics = {k: float(v) for k, v in logs.items()}
                    new_metric = valid_metrics[conf.valid_metric]
                    best_metric = valid_best_metrics.get(conf.valid_metric)
                    if best_metric is None:
                        is_new_best = True
                    elif conf.valid_higher_better:
                        is_new_best = new_metric > best_metric
                    else:
                        is_new_best = new_metric < best_metric
                    if is_new_best:
                        valid_best_metrics = deepcopy(valid_metrics)
                        save_names.append("valid_best")

                if global_step % conf.save_step == 0:
                    ckpt_dirs = [
                        key for key in os.listdir(train_dir) if key.startswith("step_")
                    ]
                    ckpt_dirs.sort(key=lambda stem: int(stem.split("_")[-1]))
                    if (
                        conf.keep_num_ckpts is not None
                        and len(ckpt_dirs) >= conf.keep_num_ckpts
                    ):
                        for ckpt_dir in ckpt_dirs[
                            : len(ckpt_dirs) - conf.keep_num_ckpts + 1
                        ]:
                            shutil.rmtree(train_dir / ckpt_dir)

                    save_names.append(f"step_{global_step}")

                for name in save_names:
                    training_stats = dict(
                        global_step=global_step,
                        epoch=epoch,
                        valid_best_metrics=valid_best_metrics,
                    )
                    _save_ckpts_to_dir(
                        train_dir / name,
                        (
                            task.module
                            if isinstance(task, _DistributedDataParallel)
                            else task
                        ),
                        optimizer,
                        scheduler,
                        build_model_all_args,
                        build_task_all_args_except_model,
                        save_model,
                        save_task,
                        training_stats,
                        global_config,
                    )

                pbar.update(1)
            epoch += 1

        pbar.close()
        if rank == 0:
            tf_logger.close()

    def evaluate(
        self,
        evaluate: dict,
        mode: str,
        task,
        dataset,
        batch_sampler,
        collate_fn,
        eval_batch: int,
        dump_dir: str,
        device: str,
        num_workers: int,
    ):
        """
        The evaluate routine used by :obj:`train` (during validation phase) and :obj:`run`
        (during testing phase).

        Args:
            evaluate (dict): same in :obj:`default_config`, no argument supported for now
            **others:
                only meaningful when you want to override this train method, which is not the
                common case. Hence we skip the documentation for now.

        """
        assert mode in ["valid", "test"]

        dataloader = DataLoader(
            dataset,
            batch_sampler=batch_sampler,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        task = task.to(device)
        with torch.no_grad():
            batch_results = []
            for batch_idx, batch in enumerate(
                tqdm(dataloader, desc=mode, total=len(dataloader))
            ):
                if batch_idx == eval_batch:
                    break
                batch = _to_device(batch, device)
                task.eval()
                loss, cacheable = task(mode, _dump_dir=dump_dir, **batch)
                batch_results.append(_force_cacheable(cacheable))

        logs = task.reduction(mode, batch_results, _dump_dir=dump_dir)
        return logs

    def save_model(
        self,
        save_model: dict,
        model_ckpt_dir: str,
        build_model_all_args: dict,
        model: torch.nn.Module,
    ):
        """
        Save the model state_dict and the model initialization arguments into the given directory.
        If you override this method, it is highly possible you also need to override :obj:`load_model`

        Args:
            save_model (dict): same in :obj:`default_config`, so the user can save additional settings,
                like the configuration of the dataset by duplicating the dataset hypers
                inside the :code:`save_model` field. You can rely on the :code:`omegaconf`
                package to simplify the duplication.
            model_ckpt_dir (str): save the model into the this directory.
            build_model_all_args (dict): all the arguments of :obj:`build_model`.
                By saving this dictionary, you can easily reconstruct the same model
                by calling :obj:`build_model` with the saved dictionary.
            model (torch.nn.Module): the model to be saved.

        Returns:
            None
        """
        model_ckpt_dir: Path = Path(model_ckpt_dir)
        if model_ckpt_dir.is_dir():
            shutil.rmtree(model_ckpt_dir, ignore_errors=True)
        model_ckpt_dir.mkdir(exist_ok=True, parents=True)

        with (model_ckpt_dir / "problem_name").open("w") as f:
            f.write(f"{self.__class__.__name__}")

        torch.save(model.state_dict(), model_ckpt_dir / "state_dict.pt")

        # NOTE: all arguments for building model should be in simple types (yaml serializable)
        with (model_ckpt_dir / f"arguments.yaml").open("w") as f:
            yaml.safe_dump(build_model_all_args, f)

        if len(save_model) > 0:
            with (model_ckpt_dir / "extra_conf.yaml").open("w") as f:
                yaml.safe_dump(save_model, f)

    def load_model(self, model_ckpt_dir: str):
        """
        Return the saved model.

        Args:
            model_ckpt_dir (str): Restore the model with :obj:`build_model` and the checkpoint
                saved in this directory.

        Return:
            :obj:`torch.nn.Module`
        """
        model_ckpt_dir: Path = Path(model_ckpt_dir)

        with (model_ckpt_dir / "arguments.yaml").open("r") as f:
            arguments = yaml.load(f, Loader=yaml.SafeLoader)
        model = self.build_model(**arguments)

        state_dict = torch.load(model_ckpt_dir / "state_dict.pt", map_location="cpu")
        model.load_state_dict(state_dict)

        return model

    def save_task(
        self,
        save_task: dict,
        task_ckpt_dir: str,
        build_task_all_args_except_model: dict,
        task: Task,
    ):
        """
        Save the task's state, :code:`task.get_state()`, and the initialization arguments into the given
        directory. If you override this method, it is highly possible you also need to override
        :obj:`load_task`.

        Args:
            save_task (dict): same in :obj:`default_config`, so the user can save additional settings,
                like the configuration of the dataset by duplicating the dataset hypers
                inside the :code:`save_task` field. You can rely on the :code:`omegaconf`
                package to simplify the duplication.
            task_ckpt_dir (str): save the task into this directory.
            build_task_all_args_except_model (dict): all the arguments of :obj:`build_task` except
                the :code:`model` argument since the model should be sapartely saved by
                :obj:`save_model`. By saving this dictionary, you can easily reconstruct the same task
                by calling :obj:`build_task` with the saved dictionary.
            task (Task): the task to be saved.

        Returns:
            None
        """
        task_ckpt_dir: Path = Path(task_ckpt_dir)
        if task_ckpt_dir.is_dir():
            shutil.rmtree(task_ckpt_dir, ignore_errors=True)
        task_ckpt_dir.mkdir(exist_ok=True, parents=True)

        with (task_ckpt_dir / "problem_name").open("w") as f:
            f.write(f"{self.__class__.__name__}")

        torch.save(task.get_state(), task_ckpt_dir / "state.pt")

        # NOTE: each argument is saved independently to prevent SPOF
        # i.e. a single argument which cannot be loaded will lead to missing
        # all other arguments, since the single argument file cannot be loaded
        arguments = build_task_all_args_except_model
        arguments_dir = task_ckpt_dir / "arguments"
        arguments_dir.mkdir(exist_ok=True, parents=True)
        for k, v in arguments.items():
            try:
                # If the object is yaml serializable, use yaml for readibility
                yaml.safe_dump(v)
            except:
                # If not, use pickle
                with (arguments_dir / f"{k}.pkl").open("wb") as f:
                    pickle.dump(v, f)
            else:
                with (arguments_dir / f"{k}.yaml").open("w") as f:
                    yaml.safe_dump(v, f)

        if len(save_task) > 0:
            with (task_ckpt_dir, "extra_conf.yaml").open("w") as f:
                yaml.safe_dump(save_task, f)

    def load_task(
        self, task_ckpt_dir: str, model: torch.nn.Module, task_overrides: dict = None
    ):
        """
        Return the saved task.

        Args:
            task_ckpt_dir (str): Restore the task with :obj:`build_task` and the checkpoint
                saved in this directory.
            model (torch.nn.Module): the model for the task, since the model is separately saved
                and is required for :obj:`build_task`.
            task_overrides (dict): overrides the saved initialization arguments, so can change
                the loaded task's behavior. Like, change the decoding hyperparameters.

        Returns:
            :obj:`s3prl.task.Task`
        """

        task_ckpt_dir: Path = Path(task_ckpt_dir)
        task_overrides = task_overrides or {}

        arguments = task_overrides.copy()
        arguments_dir = task_ckpt_dir / "arguments"
        for filename in os.listdir(arguments_dir):
            filepath = arguments_dir / filename
            key = filepath.stem

            if key in task_overrides:
                # do not load the file (potential crash) if the overrides already has the value
                continue

            if filepath.suffix == ".yaml":
                with filepath.open("r") as f:
                    value = yaml.load(f, Loader=yaml.SafeLoader)
            elif filepath.suffix == ".pkl":
                with filepath.open("rb") as f:
                    value = pickle.load(f)

            assert key not in arguments, (
                f"Unexpected duplicated file stem '{key}' found in {arguments_dir}. "
                "Please delete one of them."
            )
            arguments[key] = value

        task = self.build_task(model=model, **arguments)

        state = torch.load(Path(task_ckpt_dir) / "state.pt", map_location="cpu")
        task.set_state(state)

        return task

    def load_model_and_task(self, ckpts_dir: str, task_overrides: dict = None):
        """
        This is a helper method to combine :obj:`load_model` and :obj:`load_task`
        together to directly load the model and the task. This method assumes
        the model is saved under :code:`ckpts_dir / 'model'` and the task is
        saved under :code:`ckpts_dir / 'task'`

        Returns:
            tuple

            1. model (:obj:`torch.nn.Module`)
            2. task (:obj:`s3prl.task.Task`)
        """
        ckpts_dir: Path = Path(ckpts_dir)
        task_overrides = task_overrides or {}

        model = self.load_model(ckpts_dir / "model")
        task = self.load_task(ckpts_dir / "task", model, task_overrides)
        return model, task

    @staticmethod
    def _get_current_arguments(
        exclude_self_and_cls: bool = True, flatten_dict: Union[str, List[str]] = None
    ) -> dict:
        if isinstance(flatten_dict, str):
            flatten_dict = [flatten_dict]

        frame = inspect.currentframe().f_back
        args, _, _, values = inspect.getargvalues(frame)
        config = {key: values[key] for key in args}

        if exclude_self_and_cls:
            config.pop("self", None)
            config.pop("cls", None)

        if flatten_dict is not None:
            flatten_config = {}
            for k, v in config.items():
                if k in flatten_dict:
                    assert isinstance(v, dict)
                    for _k, _v in v.items():
                        flatten_config[_k] = _v
                else:
                    flatten_config[k] = v
            config = flatten_config

        def assert_no_missing(config: dict):
            omegaconf.OmegaConf.to_container(
                omegaconf.OmegaConf.create(config), throw_on_missing=True
            )

        assert_no_missing(config)
        return config

    @staticmethod
    def _get_time_tag():
        return datetime.fromtimestamp(time()).strftime("%Y_%m_%d_%H_%M_%S")

    @staticmethod
    def _stage_check(stage_id: int, stop: int, check_fn: callable):
        try:
            check_fn()
        except:
            logger.error(
                f"Stage {stage_id} was not done before or is corrupted. Please re-run from this stage."
            )
            raise
        if isinstance(stop, int) and stage_id == stop:
            exit(0)

    def main(self, args: List[str] = None):
        parser = argparse.ArgumentParser()
        parser.add_argument("--verbose", default="INFO")
        parser.add_argument(
            "--config", help="The yaml config path to override the default config"
        )
        parser.add_argument("--print_config", "-p", action="store_true")
        parser.add_argument(
            "--dump_config", "-d", help="The path to dump the default config as yaml"
        )
        args, override = parser.parse_known_args(args)

        if args.print_config:
            print(f"\nDefault config of {self}\n")
            print(yaml.safe_dump(self.default_config()))
            exit(0)

        if args.dump_config is not None:
            with open(args.dump_config, "w") as f:
                yaml.safe_dump(self.default_config(), f)
            exit(0)

        root_logger = logging.getLogger()
        root_logger.handlers = []
        logging.basicConfig(level=getattr(logging, args.verbose), format=LOGGING_FORMAT)

        if args.config is not None:
            with open(args.config) as f:
                yaml_conf = yaml.load(f, Loader=yaml.FullLoader) or dict()
        else:
            yaml_conf = dict()
        override_conf = parse_overrides(override)

        schema = omegaconf.OmegaConf.create(self.default_config())
        config = omegaconf.OmegaConf.merge(schema, yaml_conf, override_conf)
        config = omegaconf.OmegaConf.to_container(
            config, resolve=True, throw_on_missing=True
        )
        logger.info(config)

        self.run(**config)
        return config



"""
Diarization Permutation Invarant Task

Authors
  * Jiatong Shi 2022
  * Leo 2022
"""

from pathlib import Path
from typing import List

import torch
import torch.nn as nn

from s3prl.metric.diarization import calc_diarization_error
from s3prl.nn.pit import get_label_perm, pit_loss

from .base import Task

TOLERANT_FRAME_DIFF = 2

__all__ = ["DiarizationPIT"]


class DiarizationPIT(Task):
    def __init__(
        self,
        model: nn.Module,
    ):
        super().__init__()
        self.model = model
        self.objective = pit_loss

    def _tile_representations(self, reps, factor):
        """
        Tile up the representations by `factor`.
        Input - sequence of representations, shape: (batch_size, seq_len, feature_dim)
        Output - sequence of tiled representations, shape: (batch_size, seq_len * factor, feature_dim)
        """
        assert (
            len(reps.shape) == 3
        ), "Input argument `reps` has invalid shape: {}".format(reps.shape)
        tiled_reps = reps.repeat(1, 1, factor)
        tiled_reps = tiled_reps.reshape(
            reps.size(0), reps.size(1) * factor, reps.size(2)
        )
        return tiled_reps

    def _match_length(self, inputs, labels):
        """
        Since the upstream extraction process can sometimes cause a mismatch
        between the seq lenth of inputs and labels:
        - if len(inputs) > len(labels), we truncate the final few timestamp of inputs to match the length of labels
        - if len(inputs) < len(labels), we duplicate the last timestep of inputs to match the length of labels
        Note that the length of labels should never be changed.
        """
        input_len, label_len = inputs.size(1), labels.size(1)

        factor = int(round(label_len / input_len))
        if factor > 1:
            inputs = self._tile_representations(inputs, factor)
            input_len = inputs.size(1)

        if input_len > label_len:
            inputs = inputs[:, :label_len, :]
        elif input_len < label_len:
            pad_vec = inputs[:, -1, :].unsqueeze(1)  # (batch_size, 1, feature_dim)
            inputs = torch.cat(
                (inputs, pad_vec.repeat(1, label_len - input_len, 1)), dim=1
            )  # (batch_size, seq_len, feature_dim), where seq_len == labels.size(-1)
        return inputs, labels

    def predict(self, x, x_len):
        predicted, predicted_len = self.model(x, x_len)
        return predicted, predicted_len

    def forward(
        self,
        _mode: str,
        x,
        x_len,
        label,
        label_len,
        record_id: str,
        chunk_id: int,
        _dump_dir: str = None,
    ):
        predicted, predicted_len = self.predict(x, x_len)

        for pl, ll in zip(predicted_len, label_len):
            assert (
                abs(pl - ll) <= TOLERANT_FRAME_DIFF
            ), f"predicted: {pl}, label: {ll}, TOLERANT_FRAME_DIFF: {TOLERANT_FRAME_DIFF}"

        predicted, label = self._match_length(predicted, label)
        loss, perm_idx, perm_list = self.objective(predicted, label.float(), label_len)
        label_perm = get_label_perm(label, perm_idx, perm_list)

        (
            correct,
            num_frames,
            speech_scored,
            speech_miss,
            speech_falarm,
            speaker_scored,
            speaker_miss,
            speaker_falarm,
            speaker_error,
        ) = calc_diarization_error(predicted, label_perm, label_len)

        if speech_scored > 0 and speaker_scored > 0 and num_frames > 0:
            SAD_MR, SAD_FR, MI, FA, CF, ACC, DER = (
                speech_miss / speech_scored,
                speech_falarm / speech_scored,
                speaker_miss / speaker_scored,
                speaker_falarm / speaker_scored,
                speaker_error / speaker_scored,
                correct / num_frames,
                (speaker_miss + speaker_falarm + speaker_error) / speaker_scored,
            )
        else:
            SAD_MR, SAD_FR, MI, FA, CF, ACC, DER = 0, 0, 0, 0, 0, 0, 0

        if _mode == "test" and _dump_dir is not None:
            assert (
                len(set(list(record_id))) == 1
            ), "During testing, all utterances in a batch should come from the same recording"

            if len(label_len) > 1:
                assert (
                    len(set(label_len[:-1].tolist())) == 1
                ), f"Except the final chunk, other chunks from the same recording should have the same length"

            predicted_sorted = []
            for idx in chunk_id.long().topk(len(chunk_id), largest=False).indices:
                predicted_sorted.append(predicted[idx])

            predict = torch.vstack(predicted_sorted)
            predict = predict.detach().cpu()
            predict = 1 / (1 + (-predict).exp())

            prediction_dir = Path(_dump_dir) / f"prediction"
            prediction_dir.mkdir(exist_ok=True, parents=True)
            torch.save(predict, prediction_dir / f"{record_id[0]}.pt")

        cacheable = dict(
            loss=loss.detach().cpu(),
            accuracy=ACC,
            der=DER,
        )

        return loss, cacheable

    def reduction(self, _mode: str, cached_results: List[dict], _dump_dir: str = None):
        results = self.parse_cached_results(cached_results)
        logs = dict(
            accuracy=torch.FloatTensor(results["accuracy"]).mean().item(),
            der=torch.FloatTensor(results["der"]).mean().item(),
        )
        return logs



"""
Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py

"""
import logging
import math
import warnings
from copy import deepcopy

import torch
from torch import nn

from s3prl.util.download import urls_to_filepaths

_logger = logging.getLogger(__name__)


def adapt_input_conv(in_chans, conv_weight):
    conv_type = conv_weight.dtype
    conv_weight = (
        conv_weight.float()
    )  # Some weights are in torch.half, ensure it's float for sum on CPU
    O, I, J, K = conv_weight.shape
    if in_chans == 1:
        if I > 3:
            assert conv_weight.shape[1] % 3 == 0
            # For models with space2depth stems
            conv_weight = conv_weight.reshape(O, I // 3, 3, J, K)
            conv_weight = conv_weight.sum(dim=2, keepdim=False)
        else:
            conv_weight = conv_weight.sum(dim=1, keepdim=True)
    elif in_chans != 3:
        if I != 3:
            raise NotImplementedError("Weight format not supported by conversion.")
        else:
            # NOTE this strategy should be better than random init, but there could be other combinations of
            # the original RGB input layer weights that'd work better for specific cases.
            repeat = int(math.ceil(in_chans / 3))
            conv_weight = conv_weight.repeat(1, repeat, 1, 1)[:, :in_chans, :, :]
            conv_weight *= 3 / float(in_chans)
    conv_weight = conv_weight.to(conv_type)
    return conv_weight


def load_pretrained(
    model,
    default_cfg=None,
    num_classes=1000,
    in_chans=3,
    filter_fn=None,
    strict=True,
    progress=False,
):
    """Load pretrained checkpoint

    Args:
        model (nn.Module) : PyTorch model module
        default_cfg (Optional[Dict]): default configuration for pretrained weights / target dataset
        num_classes (int): num_classes for model
        in_chans (int): in_chans for model
        filter_fn (Optional[Callable]): state_dict filter fn for load (takes state_dict, model as args)
        strict (bool): strict load of checkpoint
        progress (bool): enable progress bar for weight download

    """
    default_cfg = default_cfg or getattr(model, "default_cfg", None) or {}
    pretrained_url = default_cfg.get("url", None)

    if not pretrained_url:
        _logger.warning(
            "No pretrained weights exist for this model. Using random initialization."
        )
        return

    _logger.info(f"Loading pretrained weights from url ({pretrained_url})")
    state_dict = torch.load(urls_to_filepaths(pretrained_url), map_location="cpu")

    if filter_fn is not None:
        # for backwards compat with filter fn that take one arg, try one first, the two
        try:
            state_dict = filter_fn(state_dict)
        except TypeError:
            state_dict = filter_fn(state_dict, model)

    input_convs = default_cfg.get("first_conv", None)
    if input_convs is not None and in_chans != 3:
        if isinstance(input_convs, str):
            input_convs = (input_convs,)
        for input_conv_name in input_convs:
            weight_name = input_conv_name + ".weight"
            try:
                state_dict[weight_name] = adapt_input_conv(
                    in_chans, state_dict[weight_name]
                )
                _logger.info(
                    f"Converted input conv {input_conv_name} pretrained weights from 3 to {in_chans} channel(s)"
                )
            except NotImplementedError as e:
                del state_dict[weight_name]
                strict = False
                _logger.warning(
                    f"Unable to convert pretrained {input_conv_name} weights, using random init for this layer."
                )

    classifiers = default_cfg.get("classifier", None)
    label_offset = default_cfg.get("label_offset", 0)
    if classifiers is not None:
        if isinstance(classifiers, str):
            classifiers = (classifiers,)
        if num_classes != default_cfg["num_classes"]:
            for classifier_name in classifiers:
                # completely discard fully connected if model num_classes doesn't match pretrained weights
                del state_dict[classifier_name + ".weight"]
                del state_dict[classifier_name + ".bias"]
            strict = False
        elif label_offset > 0:
            for classifier_name in classifiers:
                # special case for pretrained weights with an extra background class in pretrained weights
                classifier_weight = state_dict[classifier_name + ".weight"]
                state_dict[classifier_name + ".weight"] = classifier_weight[
                    label_offset:
                ]
                classifier_bias = state_dict[classifier_name + ".bias"]
                state_dict[classifier_name + ".bias"] = classifier_bias[label_offset:]

    model.load_state_dict(state_dict, strict=strict)


def overlay_external_default_cfg(default_cfg, kwargs):
    """Overlay 'external_default_cfg' in kwargs on top of default_cfg arg."""
    external_default_cfg = kwargs.pop("external_default_cfg", None)
    if external_default_cfg:
        default_cfg.pop("url", None)  # url should come from external cfg
        default_cfg.pop("hf_hub", None)  # hf hub id should come from external cfg
        default_cfg.update(external_default_cfg)


def filter_kwargs(kwargs, names):
    if not kwargs or not names:
        return
    for n in names:
        kwargs.pop(n, None)


def set_default_kwargs(kwargs, names, default_cfg):
    for n in names:
        # for legacy reasons, model __init__args uses img_size + in_chans as separate args while
        # default_cfg has one input_size=(C, H ,W) entry
        if n == "img_size":
            input_size = default_cfg.get("input_size", None)
            if input_size is not None:
                assert len(input_size) == 3
                kwargs.setdefault(n, input_size[-2:])
        elif n == "in_chans":
            input_size = default_cfg.get("input_size", None)
            if input_size is not None:
                assert len(input_size) == 3
                kwargs.setdefault(n, input_size[0])
        else:
            default_val = default_cfg.get(n, None)
            if default_val is not None:
                kwargs.setdefault(n, default_cfg[n])


def update_default_cfg_and_kwargs(default_cfg, kwargs, kwargs_filter):
    """Update the default_cfg and kwargs before passing to model

    FIXME this sequence of overlay default_cfg, set default kwargs, filter kwargs
    could/should be replaced by an improved configuration mechanism

    Args:
        default_cfg: input default_cfg (updated in-place)
        kwargs: keyword args passed to model build fn (updated in-place)
        kwargs_filter: keyword arg keys that must be removed before model __init__
    """
    # Overlay default cfg values from `external_default_cfg` if it exists in kwargs
    overlay_external_default_cfg(default_cfg, kwargs)
    # Set model __init__ args that can be determined by default_cfg (if not already passed as kwargs)
    default_kwarg_names = ("num_classes", "global_pool", "in_chans")
    if default_cfg.get("fixed_input_size", False):
        # if fixed_input_size exists and is True, model takes an img_size arg that fixes its input size
        default_kwarg_names += ("img_size",)
    set_default_kwargs(kwargs, names=default_kwarg_names, default_cfg=default_cfg)
    # Filter keyword args for task specific model variants (some 'features only' models, etc.)
    filter_kwargs(kwargs, names=kwargs_filter)


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


from torch.nn.init import _calculate_fan_in_and_fan_out


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2,
        )

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def variance_scaling_(tensor, scale=1.0, mode="fan_in", distribution="normal"):
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    if mode == "fan_in":
        denom = fan_in
    elif mode == "fan_out":
        denom = fan_out
    elif mode == "fan_avg":
        denom = (fan_in + fan_out) / 2

    variance = scale / denom

    if distribution == "truncated_normal":
        # constant is stddev of standard normal truncated to (-2, 2)
        trunc_normal_(tensor, std=math.sqrt(variance) / 0.87962566103423978)
    elif distribution == "normal":
        tensor.normal_(std=math.sqrt(variance))
    elif distribution == "uniform":
        bound = math.sqrt(3 * variance)
        tensor.uniform_(-bound, bound)
    else:
        raise ValueError(f"invalid distribution {distribution}")


def lecun_normal_(tensor):
    variance_scaling_(tensor, mode="fan_in", distribution="truncated_normal")


def build_model_with_cfg(
    model_cls,
    variant: str,
    pretrained: bool,
    default_cfg: dict,
    model_cfg=None,
    feature_cfg=None,
    pretrained_strict: bool = True,
    pretrained_filter_fn=None,
    pretrained_custom_load=False,
    kwargs_filter=None,
    **kwargs,
):
    """Build model with specified default_cfg and optional model_cfg

    This helper fn aids in the construction of a model including:
      * handling default_cfg and associated pretained weight loading
      * passing through optional model_cfg for models with config based arch spec
      * features_only model adaptation
      * pruning config / model adaptation

    Args:
        model_cls (nn.Module): model class
        variant (str): model variant name
        pretrained (bool): load pretrained weights
        default_cfg (dict): model's default pretrained/task config
        model_cfg (Optional[Dict]): model's architecture config
        feature_cfg (Optional[Dict]: feature extraction adapter config
        pretrained_strict (bool): load pretrained weights strictly
        pretrained_filter_fn (Optional[Callable]): filter callable for pretrained weights
        pretrained_custom_load (bool): use custom load fn, to load numpy or other non PyTorch weights
        kwargs_filter (Optional[Tuple]): kwargs to filter before passing to model
        **kwargs: model args passed through to model __init__
    """
    pruned = kwargs.pop("pruned", False)
    features = False
    feature_cfg = feature_cfg or {}
    default_cfg = deepcopy(default_cfg) if default_cfg else {}
    update_default_cfg_and_kwargs(default_cfg, kwargs, kwargs_filter)
    default_cfg.setdefault("architecture", variant)

    # Setup for feature extraction wrapper done at end of this fn
    if kwargs.pop("features_only", False):
        features = True
        feature_cfg.setdefault("out_indices", (0, 1, 2, 3, 4))
        if "out_indices" in kwargs:
            feature_cfg["out_indices"] = kwargs.pop("out_indices")

    # Build the model
    model = (
        model_cls(**kwargs) if model_cfg is None else model_cls(cfg=model_cfg, **kwargs)
    )
    model.default_cfg = default_cfg

    # For classification models, check class attr, then kwargs, then default to 1k, otherwise 0 for feats
    num_classes_pretrained = (
        0
        if features
        else getattr(model, "num_classes", kwargs.get("num_classes", 1000))
    )
    if pretrained:
        assert not pretrained_custom_load, "URL should not contain npz for PASST models"
        load_pretrained(
            model,
            num_classes=num_classes_pretrained,
            in_chans=kwargs.get("in_chans", 3),
            filter_fn=pretrained_filter_fn,
            strict=pretrained_strict,
        )
    return model



from s3prl.metric import cer, per, wer


def isclose(x: float, y: float) -> float:
    return abs(x - y) < 1e-9


def test_metric():
    # test wer & cer
    hyps = ["a ac abb d"]
    refs = ["a ab abc d"]

    assert isclose(cer(hyps, refs), 0.2)
    assert isclose(wer(hyps, refs), 0.5)
    assert isclose(per(hyps, refs), 0.5)



"""Largely inspired by https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/vit.py"""
import torch
import torch.nn as nn
from einops import repeat
from einops.layers.torch import Rearrange


class SST(nn.Module):
    def __init__(
        self,
        image_size,
        patch_size,
        embed_dim,
        depth,
        num_heads,
        channels=1,
        dropout=0.1,
        emb_dropout=0.1,
    ):
        super().__init__()
        image_height, image_width = image_size
        patch_height, patch_width = patch_size

        assert (
            image_height % patch_height == 0 and image_width % patch_width == 0
        ), "Image dimensions must be divisible by the patch size."

        num_patches = (image_height // patch_height) * (image_width // patch_width)
        patch_dim = channels * patch_height * patch_width
        # assert pool in {'cls', 'mean'}, 'pool type must be either cls (cls token) or mean (mean pooling)'

        self.to_patch_embedding = nn.Sequential(
            Rearrange(
                "b c (h p1) (w p2) -> b (h w) (p1 p2 c)",
                p1=patch_height,
                p2=patch_width,
            ),
            nn.Linear(patch_dim, embed_dim),
        )

        self.pos_embedding = nn.Parameter(
            torch.randn(1, num_patches, embed_dim)
        )  # Learnable embedding
        # self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, embed_dim))
        # self.cls_token = nn.Parameter(torch.randn(1, 1, embed_dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=embed_dim,
            nhead=num_heads,
            dim_feedforward=embed_dim * num_heads,
            dropout=dropout,
            activation="gelu",
        )

        self.transformer = nn.TransformerEncoder(self.encoder_layer, num_layers=depth)

    def forward(self, img):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        # No need for cls_token? We're not doing classification
        # cls_tokens = repeat(self.cls_token, '() n d -> b n d', b=b)
        # x = torch.cat((cls_tokens, x), dim=1)

        x += self.pos_embedding[:, : (n + 1)]

        x = self.dropout(x)

        x = self.transformer(x)

        # x = x.mean(1)  # https://github.com/lucidrains/vit-pytorch/blob/main/vit_pytorch/vit.py

        x = x.mean(1) + x.amax(1)  # BYOL-A mean + max

        return x



"""
Common pooling methods

Authors:
  * Leo 2022
  * Haibin Wu 2022
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = [
    "MeanPooling",
    "TemporalAveragePooling",
    "TemporalStatisticsPooling",
    "SelfAttentivePooling",
    "AttentiveStatisticsPooling",
]


class MeanPooling(nn.Module):
    """
    Computes Temporal Average Pooling (MeanPooling over time) Module
    """

    def __init__(self, input_size: int):
        super().__init__()
        self._in_size = input_size

    @property
    def input_size(self) -> int:
        return self._in_size

    @property
    def output_size(self) -> int:
        return self._in_size

    def forward(self, xs: torch.Tensor, xs_len: torch.LongTensor):
        """
        Args:
            xs (torch.Tensor): Input tensor (#batch, frames, input_size).
            xs_len (torch.LongTensor): with the lengths for each sample
        Returns:
            torch.Tensor: Output tensor (#batch, input_size)
        """
        pooled_list = []
        for x, x_len in zip(xs, xs_len):
            pooled = torch.mean(x[:x_len], dim=0)
            pooled_list.append(pooled)
        return torch.stack(pooled_list)


TemporalAveragePooling = MeanPooling


class TemporalStatisticsPooling(nn.Module):
    """
    TemporalStatisticsPooling
    Paper: X-vectors: Robust DNN Embeddings for Speaker Recognition
    Link: http://www.danielpovey.com/files/2018_icassp_xvectors.pdf
    """

    def __init__(self, input_size: int):
        super().__init__()
        self._input_size = input_size

    @property
    def input_size(self) -> int:
        return self._input_size

    @property
    def output_size(self) -> int:
        return self._input_size * 2

    def forward(self, xs, xs_len):
        """
        Computes Temporal Statistics Pooling Module

        Args:
            xs (torch.Tensor): Input tensor (#batch, frames, input_size).
            xs_len (torch.LongTensor): with the lengths for each sample

        Returns:
            torch.Tensor: Output tensor (#batch, output_size)
        """
        pooled_list = []
        for x, x_len in zip(xs, xs_len):
            mean = torch.mean(x[:x_len], dim=0)
            std = torch.std(x[:x_len], dim=0)
            pooled = torch.cat((mean, std), dim=-1)
            pooled_list.append(pooled)
        return torch.stack(pooled_list)


class SelfAttentivePooling(nn.Module):
    """
    SelfAttentivePooling
    Paper: Self-Attentive Speaker Embeddings for Text-Independent Speaker Verification
    Link: https://danielpovey.com/files/2018_interspeech_xvector_attention.pdf
    """

    def __init__(self, input_size: int):
        super().__init__()
        self._indim = input_size
        self.sap_linear = nn.Linear(input_size, input_size)
        self.attention = nn.Parameter(torch.FloatTensor(input_size, 1))

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._indim

    def forward(self, xs, xs_len):
        """
        Computes Self-Attentive Pooling Module

        Args:
            xs (torch.Tensor): Input tensor (#batch, frames, input_size).
            xs_len (torch.LongTensor): with the lengths for each sample

        Returns:
            torch.Tensor: Output tensor (#batch, input_size)
        """
        pooled_list = []
        for x, x_len in zip(xs, xs_len):
            x = x[:x_len].unsqueeze(0)
            h = torch.tanh(self.sap_linear(x))
            w = torch.matmul(h, self.attention).squeeze(dim=2)
            w = F.softmax(w, dim=1).view(x.size(0), x.size(1), 1)
            x = torch.sum(x * w, dim=1)
            pooled_list.append(x.squeeze(0))
        return torch.stack(pooled_list)


class AttentiveStatisticsPooling(nn.Module):
    """
    AttentiveStatisticsPooling
    Paper: Attentive Statistics Pooling for Deep Speaker Embedding
    Link: https://arxiv.org/pdf/1803.10963.pdf
    """

    def __init__(self, input_size: int):
        super().__init__()
        self._indim = input_size
        self.sap_linear = nn.Linear(input_size, input_size)
        self.attention = nn.Parameter(torch.FloatTensor(input_size, 1))

    @property
    def input_size(self) -> int:
        return self._indim

    @property
    def output_size(self) -> int:
        return self._indim * 2

    def forward(self, xs, xs_len):
        """
        Computes Attentive Statistics Pooling Module

        Args:
            xs (torch.Tensor): Input tensor (#batch, frames, input_size).
            xs_len (torch.LongTensor): with the lengths for each sample

        Returns:
            torch.Tensor: Output tensor (#batch, input_size)
        """
        pooled_list = []
        for x, x_len in zip(xs, xs_len):
            x = x[:x_len].unsqueeze(0)
            h = torch.tanh(self.sap_linear(x))
            w = torch.matmul(h, self.attention).squeeze(dim=2)
            w = F.softmax(w, dim=1).view(x.size(0), x.size(1), 1)
            mu = torch.sum(x * w, dim=1)
            rh = torch.sqrt((torch.sum((x**2) * w, dim=1) - mu**2).clamp(min=1e-5))
            x = torch.cat((mu, rh), 1).squeeze(0)
            pooled_list.append(x)
        return torch.stack(pooled_list)



# Copyright (c) Facebook, Inc. All Rights Reserved

# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/multires_hubert/hubconf.py ]
#   Synopsis     [ the Multiresolution HuBERT torch hubconf ]
#   Author       [ S3PRL / Jiatong Shi ]
"""*********************************************************************************************"""

# isort: off
import logging
import os
import time
from pathlib import Path

from filelock import FileLock
from s3prl.util.download import _urls_to_filepaths

from .convert import load_and_convert_fairseq_ckpt
from .expert import UpstreamExpert as _UpstreamExpert

# isort: on

logger = logging.getLogger(__name__)

NEW_ENOUGH_SECS = 2.0


def multires_hubert_custom(
    ckpt: str,
    refresh: bool = False,
    **kwargs,
):

    if ckpt.startswith("http"):
        ckpt = _urls_to_filepaths(ckpt, refresh=refresh)
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt=ckpt, **kwargs)


def multires_hubert_local(*args, **kwargs):
    return multires_hubert_custom(*args, **kwargs)


def multires_hubert_base(refresh=False, **kwargs):
    """
    The monolingual base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/mr_hubert/resolve/main/mrhubert_mono_base.pt"
    return multires_hubert_custom(refresh=refresh, **kwargs)


def multires_hubert_large(refresh=False, **kwargs):
    """
    The monolingual base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/mr_hubert/resolve/main/mrhubert_mono_large.pt"
    return multires_hubert_custom(refresh=refresh, **kwargs)


def multires_hubert_multilingual_base(refresh=False, **kwargs):
    """
    The multilingual base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://huggingface.co/s3prl/mr_hubert/resolve/main/multi_base.pt"
    return multires_hubert_custom(refresh=refresh, **kwargs)


def multires_hubert_multilingual_large400k(refresh=False, **kwargs):
    """
    The multilingual large model (400k steps)
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/mr_hubert/resolve/main/multi_large_400k.pt"
    return multires_hubert_custom(refresh=refresh, **kwargs)


def multires_hubert_multilingual_large600k(refresh=False, **kwargs):
    """
    The multilingual large model (600k steps)
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/mr_hubert/resolve/main/multi_large_600k.pt"
    return multires_hubert_custom(refresh=refresh, **kwargs)



"""
Parse command-line arguments into override dictionary

Authors
  * Leo 2022
"""

import logging

logger = logging.getLogger(__name__)

__all__ = [
    "parse_overrides",
]


def parse_override(string):
    """
    Example usgae:
        -o "optimizer.lr=1.0e-3,,optimizer.name='AdamW',,runner.eval_dataloaders=['dev', 'test']"

    Convert to:
        {
            "optimizer": {"lr": 1.0e-3, "name": "AdamW"},
            "runner": {"eval_dataloaders": ["dev", "test"]}
        }
    """
    options = string.split(",,")
    config = {}
    for option in options:
        option = option.strip()
        key, value_str = option.split("=")
        key, value_str = key.strip(), value_str.strip()
        remaining = key.split(".")

        try:
            value = eval(value_str)
        except:
            value = value_str

        logger.info(f"{key} = {value}")

        target_config = config
        for i, field_name in enumerate(remaining):
            if i == len(remaining) - 1:
                target_config[field_name] = value
            else:
                target_config.setdefault(field_name, {})
                target_config = target_config[field_name]
    return config


def parse_overrides(options: list):
    """
    Example usgae:
        [
            "--optimizer.lr",
            "1.0e-3",
            "--optimizer.name",
            "AdamW",
            "--runner.eval_dataloaders",
            "['dev', 'test']",
        ]

    Convert to:
        {
            "optimizer": {"lr": 1.0e-3, "name": "AdamW"},
            "runner": {"eval_dataloaders": ["dev", "test"]}
        }
    """
    config = {}
    for position in range(0, len(options), 2):
        key: str = options[position]
        assert key.startswith("--")
        key = key.strip("--")
        value_str: str = options[position + 1]
        key, value_str = key.strip(), value_str.strip()
        remaining = key.split(".")

        try:
            value = eval(value_str)
        except Exception as e:
            if "newdict" in value_str or "Container" in value_str:
                raise
            value = value_str

        logger.debug(f"{key} = {value}")

        target_config = config
        for i, field_name in enumerate(remaining):
            if i == len(remaining) - 1:
                target_config[field_name] = value
            else:
                target_config.setdefault(field_name, {})
                target_config = target_config[field_name]
    return config



import importlib
from typing import List

import torch
import torch.nn as nn
import torchaudio
from torch.nn.utils.rnn import pad_sequence

SAMPLE_RATE = 16000


class UpstreamExpert(nn.Module):
    def __init__(
        self,
        name: str,
        refresh=False,
        window_secs: float = 0.16,
        stride_secs: float = 0.05,
    ):
        super().__init__()
        self.resampler = torchaudio.transforms.Resample(16000, 32000)
        self.module = importlib.import_module(f".hear21passt.{name}", __package__)
        self.model = self.module.load_model(
            timestamp_window=window_secs * 1000,
            timestamp_hop=stride_secs * 1000,
        )
        self.stride_secs = stride_secs

    def get_downsample_rates(self, key=None):
        return int(self.stride_secs * SAMPLE_RATE)

    def forward(self, wavs: List[torch.Tensor]):
        wavs = pad_sequence(wavs, batch_first=True)
        wavs = self.resampler(wavs)
        embs, timestamps = self.module.get_timestamp_embeddings(wavs, self.model)

        return {
            "hidden_states": [embs],
        }



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/byol_a/expert.py ]
#   Synopsis     [ the BYOL-Audio wrapper ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import logging

import torch
import torch.nn as nn
import torchaudio

from .byol_a import AudioNTT2020, PrecomputedNorm, load_yaml_config

logger = logging.getLogger(__name__)

SAMPLE_RATE = 16000


class UpstreamExpert(nn.Module):
    """
    The BYOL-A wrapper
    """

    def __init__(
        self,
        ckpt: str,
        model_config: str,
        feature_d: int,
        window_secs: float = 1.0,
        stride_secs: float = 1.0,
    ):
        super().__init__()
        config = load_yaml_config(model_config)

        self.window_secs = window_secs
        self.stride_secs = stride_secs
        self.output_dim = feature_d
        self.seg_input_length = len(
            range(0, int(window_secs * SAMPLE_RATE), config.hop_length)
        )

        # Preprocessor and normalizer.
        self.to_melspec = torchaudio.transforms.MelSpectrogram(
            sample_rate=config.sample_rate,
            n_fft=config.n_fft,
            win_length=config.win_length,
            hop_length=config.hop_length,
            n_mels=config.n_mels,
            f_min=config.f_min,
            f_max=config.f_max,
        )
        stats = [
            -5.4919195,
            5.0389895,
        ]  # FIXME: should use downstream dataset statistics
        self.normalizer = PrecomputedNorm(stats)

        # Load pretrained weights.
        self.model = AudioNTT2020(d=feature_d)
        self.model.load_weight(ckpt, device="cpu")

    def get_downsample_rates(self, key: str = None) -> int:
        return int(self.stride_secs * SAMPLE_RATE)

    def forward(self, wavs):
        """
        Args:
            wavs:
                list of unpadded wavs [wav1, wav2, ...]
                each wav is in torch.FloatTensor with sample rate 16000
                and already put in the device assigned by command-line args

        Return:
            features:
                list of unpadded features [feat1, feat2, ...]
                each feat is in torch.FloatTensor and already
                put in the device assigned by command-line args
        """
        wavs_len = [len(wav) for wav in wavs]
        max_wav_len = max(wavs_len)
        start_points = list(range(0, max_wav_len, int(self.stride_secs * SAMPLE_RATE)))
        padded_max_wav_len = start_points[-1] + int(self.window_secs * SAMPLE_RATE)
        padded_wavs = [
            torch.cat([wav, wav.new_zeros(padded_max_wav_len - len(wav))])
            for wav in wavs
        ]

        all_features = []
        for start in start_points:
            subwavs = [
                wav[start : start + int(self.window_secs * SAMPLE_RATE)]
                for wav in padded_wavs
            ]
            features = [
                self.normalizer(
                    (self.to_melspec(wav) + torch.finfo(torch.float).eps).log()
                ).permute(1, 0)
                for wav in subwavs
            ]
            features = torch.stack(
                features, dim=0
            )  # (batch_size, segment_seq_len, hiddqen_size)
            all_features.append(features)

        all_features = torch.stack(all_features, dim=0)
        num_segment, batch_size, segment_seq_len, hidden_size = all_features.shape

        flatten_features = all_features.reshape(-1, segment_seq_len, hidden_size)

        repre = self.model(
            flatten_features.transpose(1, 2).unsqueeze(1)
        )  # repre: (num_segment * batch_size, hidden_size)
        repre = repre.reshape(num_segment, batch_size, -1).transpose(
            0, 1
        )  # repre: (batch_size, num_segment, hidden_size)

        trimmed_hs = []
        for h in [repre]:
            max_h_len = len(range(0, max_wav_len, self.get_downsample_rates()))
            h = h[:, :max_h_len, :]
            trimmed_hs.append(h)

        return {
            "hidden_states": trimmed_hs,
        }



import copy
import random
from dataclasses import dataclass
from functools import lru_cache

import numpy as np
import torch

from .base import AugmentedDynamicItemDataset, DataPipe

MAX_SEQLEN = 10000


@dataclass
class PrepareTargetFeat(DataPipe):
    use_copy: bool = True
    source_feat_name: str = "source_feat"
    target_feat_name: str = "target_feat"

    def prepare_target_feat(self, feat):
        target_feat = copy.deepcopy(feat) if self.use_copy else feat
        return target_feat.to(dtype=torch.float32)

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.prepare_target_feat,
            takes=self.source_feat_name,
            provides=self.target_feat_name,
        )
        return dataset


@dataclass
class MaskedReconstruction(DataPipe):
    position_encoding_size: int = 768
    mask_proportion: float = 0.15
    mask_consecutive_min: int = 7
    mask_consecutive_max: int = 7
    mask_allow_overlap: bool = True
    mask_bucket_ratio: float = 1.5
    mask_frequency: int = 0
    source_feat_name: str = "source_feat"
    target_feat_name: str = "target_feat"
    masked_feat_name: str = "masked_feat"
    pos_enc_name: str = "pos_enc"
    attn_mask_name: str = "attn_mask"
    label_mask_name: str = "label_mask"
    """
    Args:
        position_encoding_size (int): this should be identical to `hidden_size`
        mask_proportion (float): mask this percentage of all spectrogram frames in each sequence at random during MAM training
        mask_consecutive_min (int): mask this amount of consecutive frames
        mask_consecutive_max (int): mask this amount of consecutive frames
        mask_allow_overlap (bool): allow overlap masking
        mask_bucket_ratio (float): only used when overlap is not allowed. sample a mask from each bucket in size of [sampled mask_consecutive * mask_bucket_ratio]
        mask_frequency (float): mask maximum this percentage of frequency bands, set to 0 for no frequency mask
        source_feat_name (str): handle for the `takes` (input)
        target_feat_name (str): handle for the `takes` (input)
        masked_feat_name (str): handle for the `provides` (output)
        pos_enc_name (str): handle for the `provides` (output)
        attn_mask_name (str): handle for the `provides` (output)
        label_mask_name (str): handle for the `provides` (output)
    """

    def generate_masked_data(self, source_feat, target_feat):
        with torch.no_grad():
            masked_feat = copy.deepcopy(source_feat)

            # Record length for each uttr
            spec_len = (target_feat.sum(dim=-1) != 0).long().sum(dim=-1).tolist()
            seq_len = target_feat.shape[0]

            pos_enc = fast_position_encoding(
                seq_len, self.position_encoding_size
            )  # (seq_len, position_encoding_size)
            label_mask = (
                torch.zeros_like(target_feat, dtype=torch.uint8)
                if self.mask_proportion != 0 or self.mask_frequency != 0
                else torch.ones_like(target_feat, dtype=torch.uint8)
            )
            attn_mask = torch.ones(seq_len)  # (seq_len)

            # zero vectors for padding dimension
            attn_mask[spec_len:] = 0

            def _starts_to_intervals(starts, consecutive):
                tiled = starts.expand(consecutive, starts.size(0)).permute(1, 0)
                offset = torch.arange(consecutive).expand_as(tiled)
                intervals = tiled + offset
                return intervals.view(-1)

            # time masking
            if self.mask_proportion > 0:
                mask_consecutive = random.randint(
                    self.mask_consecutive_min,
                    self.mask_consecutive_max,
                )
                valid_start_max = max(
                    spec_len - mask_consecutive - 1, 0
                )  # compute max valid start point for a consecutive mask
                proportion = round(spec_len * self.mask_proportion / mask_consecutive)
                if self.mask_allow_overlap:
                    # draw `proportion` samples from the range (0, valid_index_range) and without replacement
                    chosen_starts = torch.randperm(valid_start_max + 1)[:proportion]
                else:
                    mask_bucket_size = round(mask_consecutive * self.mask_bucket_ratio)
                    rand_start = random.randint(
                        0, min(mask_consecutive, valid_start_max)
                    )
                    valid_starts = torch.arange(
                        rand_start, valid_start_max + 1, mask_bucket_size
                    )
                    chosen_starts = valid_starts[
                        torch.randperm(len(valid_starts))[:proportion]
                    ]
                chosen_intervals = _starts_to_intervals(chosen_starts, mask_consecutive)

                # determine whether to mask / random / or do nothing to the frame
                dice = random.random()
                # mask to zero
                if dice < 0.8:
                    masked_feat[chosen_intervals, :] = 0
                # replace to random frames
                elif dice >= 0.8 and dice < 0.9:
                    random_starts = torch.randperm(valid_start_max + 1)[:proportion]
                    random_intervals = _starts_to_intervals(
                        random_starts, mask_consecutive
                    )
                    masked_feat[chosen_intervals, :] = masked_feat[random_intervals, :]
                # do nothing
                else:
                    pass

                # the gradients will be calculated on chosen frames
                label_mask[chosen_intervals, :] = 1

            # frequency masking
            if self.mask_frequency > 0:
                max_width = int(masked_feat.shape[-1] * self.mask_frequency)
                rand_bandwidth = random.randint(0, max_width)
                chosen_starts = torch.randperm(masked_feat.shape[-1] - rand_bandwidth)[
                    :1
                ]
                chosen_intervals = _starts_to_intervals(chosen_starts, rand_bandwidth)
                masked_feat[:, chosen_intervals] = 0

                # the gradients will be calculated on chosen frames
                label_mask[:spec_len, chosen_intervals] = 1

            masked_feat = masked_feat.to(dtype=torch.float32)
            pos_enc = pos_enc.to(dtype=torch.float32)
            attn_mask = attn_mask.to(dtype=torch.float32)
            label_mask = label_mask.to(dtype=torch.bool)

        return masked_feat, pos_enc, attn_mask, label_mask

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.generate_masked_data,
            takes=[self.source_feat_name, self.target_feat_name],
            provides=[
                self.masked_feat_name,
                self.pos_enc_name,
                self.attn_mask_name,
                self.label_mask_name,
            ],
        )
        return dataset


@lru_cache(maxsize=128)
def get_sinusoid_table(hidden_size):
    def cal_angle(position, hid_idx):
        return position / np.power(10000, 2 * (hid_idx // 2) / hidden_size)

    def get_posi_angle_vec(position):
        return [cal_angle(position, hid_j) for hid_j in range(hidden_size)]

    sinusoid_table = np.array(
        [get_posi_angle_vec(pos_i) for pos_i in range(MAX_SEQLEN)]
    )
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1
    return torch.FloatTensor(sinusoid_table)


def fast_position_encoding(seq_len, hidden_size, batch_size=None, padding_idx=None):
    assert (
        seq_len <= MAX_SEQLEN
    ), f"constant MAX_SEQLEN ({MAX_SEQLEN}) in mam.py < received seq_len ({seq_len})"
    table = get_sinusoid_table(hidden_size)[:seq_len]

    if padding_idx is not None:
        # deepcopy will slow down whole process when positional table is too large
        # this path is dreprecated and should never be used
        table = copy.deepcopy(table)
        table[padding_idx:] = 0.0  # zero vector for padding dimension

    if batch_size is not None:
        # using expand will not cause extra CPU memory allocation issue
        # however, the expanded tensor after put into GPU still need
        # GPU memory of expanded size, which should be avoided when
        # positional table is large
        # this path is not recommended
        batch_table = table.expand(batch_size, -1, -1)
        return batch_table  # (batch_size, seq_len, hidden_size)
    else:
        # this path is most recommended, no extra CPU and GPU memory allocation
        # after getting the (seq_len, hidden_size) tensor, one should first put
        # this tensor into GPU then expand it
        return table  # (seq_len, hidden_size)



from .base import SequentialDataPipe
from .common_pipes import LoadAudio, Phonemize, SetOutputKeys


class Speech2PhonemePipe(SequentialDataPipe):
    """
    each item in the input dataset should have:
        wav_path: str
        transcription: str
    """

    def __init__(self):
        output_keys = dict(
            x="wav",
            x_len="wav_len",
            labels="phonemized_text",
            class_ids="tokenized_text",
            unique_name="id",
        )

        super().__init__(
            LoadAudio(),
            Phonemize(),
            SetOutputKeys(output_keys=output_keys),
        )
